# import standard libraries
import logging
from importlib.resources import files
from pathlib import Path
from shutil import copyfile
from typing import Any, Dict, List, Optional, Set, Tuple, Union
# import local files
from ogd.core.schemas.Schema import Schema
from ogd.core.schemas.games.AggregateSchema import AggregateSchema
from ogd.core.schemas.games.DetectorSchema import DetectorSchema
from ogd.core.schemas.games.DetectorMapSchema import DetectorMapSchema
from ogd.core.schemas.games.DataElementSchema import DataElementSchema
from ogd.core.schemas.games.EventSchema import EventSchema
from ogd.core.schemas.games.PerCountSchema import PerCountSchema
from ogd.core.schemas.games.FeatureSchema import FeatureSchema
from ogd.core.schemas.games.FeatureMapSchema import FeatureMapSchema
from ogd.core.models.enums.IterationMode import IterationMode
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.utils import utils
from ogd.core.utils.utils import loadJSONFile
from ogd.core.utils.Logger import Logger

## @class GameSchema
#  A fairly simple class that reads a JSON schema with information on how a given
#  game's data is structured in the database, and the features we want to extract
#  for that game.
#  The class includes several functions for easy access to the various parts of
#  this schema data.
class GameSchema(Schema):
    # *** BUILT-INS & PROPERTIES ***

    # TODO: need to get game_state from schema file.

    def __init__(self, name:str, all_elements:Dict[str, Any]):
        """Constructor for the GameSchema class.
        Given a path and filename, it loads the data from a JSON schema,
        storing the full schema into a private variable, and compiling a list of
        all features to be extracted.

        :param schema_name: The name of the JSON schema file (if .json is not the file extension, .json will be appended)
        :type schema_name: str
        :param schema_path: schema_path Path to the folder containing the JSON schema file; if None is given, defaults to ./ogd/games/{game_id}/schemas/
        :type schema_path: str, optional
        """
    # 1. define instance vars
        self._game_id                : str                                  = name
        self._enum_defs              : Dict[str, List[str]]                 = {}
        self._game_state             : Dict[str, Any]                       = {}
        self._user_data              : Dict[str, Any]                       = {}
        self._event_list             : List[EventSchema]                    = []
        self._detector_map           : Dict[str, Dict[str, DetectorSchema]] = {'perlevel':{}, 'per_count':{}, 'aggregate':{}}
        self._aggregate_feats        : Dict[str, AggregateSchema]           = {}
        self._percount_feats         : Dict[str, PerCountSchema]            = {}
        self._legacy_perlevel_feats  : Dict[str, PerCountSchema]            = {}
        self._legacy_mode            : bool                                 = False
        self._config                 : Dict[str, Any]                       = {}
        self._min_level              : Optional[int]                        = None
        self._max_level              : Optional[int]                        = None
        self._other_ranges           : Dict[str, range]
        self._supported_vers         : Optional[List[int]]

        if not isinstance(all_elements, dict):
            all_elements   = {}
            Logger.Log(f"For {self._game_id} GameSchema, all_elements was not a dict, defaulting to empty dict", logging.WARN)

    # 2. set instance vars, starting with event data

        if "enums" in all_elements.keys():
            self._enum_defs = GameSchema._parseEnumDefs(enums_list=all_elements['enums'])
        else:
            Logger.Log(f"{self._game_id} game schema does not specify any custom enums", logging.INFO)

        if "game_state" in all_elements.keys():
            self._game_state = GameSchema._parseGameState(game_state=all_elements['game_state'])
        else:
            Logger.Log(f"{self._game_id} game schema does not specify the structure of the game state colum, defaulting to empty dictn", logging.INFO)

        if "user_data" in all_elements.keys():
            self._user_data = GameSchema._parseUserData(user_data=all_elements['user_data'])
        else:
            Logger.Log(f"{self._game_id} game schema does not specify the structure of the user data colum, defaulting to empty dictn", logging.INFO)
        if "events" in all_elements.keys():
            self._event_list = GameSchema._parseEventList(events_list=all_elements['events'])
        else:
            Logger.Log(f"{self._game_id} game schema does not document any events.", logging.INFO)

    # 3. Get detector information
        if "detectors" in all_elements.keys():
            # TODO : Just have DetectorMapSchema directly
            _detector_map = GameSchema._parseDetectorMap(detector_map=all_elements['detectors'])
            self._detector_map = _detector_map.AsDict
        else:
            Logger.Log(f"{self._game_id} game schema does not define any detectors.", logging.INFO)

    # 4. Get feature information
        if "features" in all_elements.keys():
            # TODO : Just have the FeatureMapSchema directly, not 4 different things.
            _feat_map = GameSchema._parseFeatureMap(feature_map=all_elements['features'])
            self._aggregate_feats.update(_feat_map.AggregateFeatures)
            self._percount_feats.update(_feat_map.PerCountFeatures)
            self._legacy_perlevel_feats.update(_feat_map.LegacyPerLevelFeatures)
            self._legacy_mode = _feat_map.LegacyMode
        else:
            Logger.Log(f"{self._game_id} game schema does not define any features.", logging.INFO)

    # 5. Get config, if any
        if "config" in all_elements.keys():
            self._config = all_elements['config']
        else:
            Logger.Log(f"{self._game_id} game schema does not define any config items.", logging.INFO)
        if "SUPPORTED_VERS" in self._config:
            self._supported_vers = self._config['SUPPORTED_VERS']
        else:
            self._supported_vers = None
            Logger.Log(f"{self._game_id} game schema does not define supported versions, defaulting to support all versions.", logging.INFO)

    # 6. Get level range and other ranges, if any
        if "level_range" in all_elements.keys():
            self._min_level, self._max_level = GameSchema._parseLevelRange(all_elements['level_range'])
        else:
            Logger.Log(f"{self._game_id} game schema does not define a level range.", logging.INFO)

        self._other_ranges = {key : range(val.get('min', 0), val.get('max', 1)) for key,val in all_elements.items() if key.endswith("_range")}

    # 7. Collect any other, unexpected elements
        _leftovers = { key:val for key,val in all_elements.items() if key not in {'enums', 'game_state', 'user_data', 'events', 'detectors', 'features', 'level_range', 'config'}.union(self._other_ranges.keys()) }
        super().__init__(name=self._game_id, other_elements=_leftovers)

    # *** BUILT-INS & PROPERTIES ***

    # def __getitem__(self, key) -> Any:
    #     return _schema[key] if _schema is not None else None

    @property
    def GameName(self) -> str:
        """Property for the name of the game configured by this schema
        """
        return self._game_id

    @property
    def EnumDefs(self) -> Dict[str, List[str]]:
        """Property for the dict of all enums defined for sub-elements in the given game's schema.
        """
        return self._enum_defs

    @property
    def GameState(self) -> Dict[str, Any]:
        """Property for the dictionary describing the structure of the GameState column for the given game.
        """
        return self._game_state

    @property
    def UserData(self) -> Dict[str, Any]:
        """Property for the dictionary describing the structure of the UserData column for the given game.
        """
        return self._user_data

    @property
    def Events(self) -> List[EventSchema]:
        """Property for the list of events the game logs.
        """
        return self._event_list

    @property
    def EventTypes(self) -> List[str]:
        """Property for the names of all event types for the game.
        """
        return [event.Name for event in self.Events]

    @property
    def Detectors(self) -> Dict[str, Dict[str, DetectorSchema]]:
        """Property for the dictionary of categorized detectors to extract.
        """
        return self._detector_map

    @property
    def DetectorNames(self) -> List[str]:
        """Property for the compiled list of all detector names.
        """
        ret_val : List[str] = []
        for _category in self.Detectors.values():
            ret_val += [detector.Name for detector in _category.values()]
        return ret_val

    @property
    def PerCountDetectors(self) -> Dict[str, DetectorSchema]:
        """Property for the dictionary of per-custom-count detectors.
        """
        return self.Detectors.get("per_count", {})

    @property
    def AggregateDetectors(self) -> Dict[str, DetectorSchema]:
        """Property for the dictionary of aggregate detectors.
        """
        return self.Detectors.get("aggregate", {})

    @property
    def Features(self) -> Dict[str, Union[Dict[str, AggregateSchema], Dict[str, PerCountSchema]]]:
        """Property for the dictionary of categorized features to extract.
        """
        return { 'aggregate' : self._aggregate_feats, 'per_count' : self._percount_feats, 'perlevel' : self._legacy_perlevel_feats }

    @property
    def FeatureNames(self) -> List[str]:
        """Property for the compiled list of all feature names.
        """
        ret_val : List[str] = []
        for _category in self.Features.values():
            ret_val += [feature.Name for feature in _category.values()]
        return ret_val

    @property
    def LegacyPerLevelFeatures(self) -> Dict[str,PerCountSchema]:
        """Property for the dictionary of legacy per-level features
        """
        return self._legacy_perlevel_feats

    @property
    def PerCountFeatures(self) -> Dict[str,PerCountSchema]:
        """Property for the dictionary of per-custom-count features.
        """
        return self._percount_feats

    @property
    def AggregateFeatures(self) -> Dict[str,AggregateSchema]:
        """Property for the dictionary of aggregate features.
        """
        return self._aggregate_feats

    @property
    def LevelRange(self) -> range:
        """Property for the range of levels defined in the schema if any.
        """
        ret_val = range(0)
        if self._min_level is not None and self._max_level is not None:
            # for i in range(self._min_level, self._max_level+1):
            ret_val = range(self._min_level, self._max_level+1)
        else:
            Logger.Log(f"Could not generate per-level features, min_level={self._min_level} and max_level={self._max_level}", logging.ERROR)
        return ret_val

    @property
    def OtherRanges(self) -> Dict[str, range]:
        return self._other_ranges

    @property
    def Config(self) -> Dict[str, Any]:
        return self._config

    @property
    def AsMarkdown(self) -> str:
        event_summary = ["## Logged Events",
                         "The individual fields encoded in the *game_state* and *user_data* Event element for all event types, and the fields in the *event_data* Event element for each individual event type logged by the game."
                        ]
        enum_list     = ["### Enums",
                         "\n".join(
                             ["| **Name** | **Values** |",
                             "| ---      | ---        |"]
                         + [f"| {name} | {val_list} |" for name,val_list in self.EnumDefs.items()]
                        )]
        game_state_list = ["### Game State",
                           "\n".join(
                               ["| **Name** | **Type** | **Description** | **Sub-Elements** |",
                               "| ---      | ---      | ---             | ---         |"]
                           + [elem.AsMarkdownRow for elem in self.GameState.values()]
                          )]
        user_data_list = ["### User Data",
                          "\n".join(
                              ["| **Name** | **Type** | **Description** | **Sub-Elements** |",
                              "| ---      | ---      | ---             | ---         |"]
                          + [elem.AsMarkdownRow for elem in self.UserData.values()]
                         )]
        # Set up list of events
        event_list = [event.AsMarkdownTable for event in self.Events] if len(self.Events) > 0 else ["None"]
        # Set up list of detectors
        detector_summary = ["## Detected Events",
                            "The custom, data-driven Events calculated from this game's logged events by OpenGameData when an 'export' is run."
                           ]
        detector_list = []
        for detect_kind in ["perlevel", "per_count", "aggregate"]:
            if detect_kind in self._detector_map:
                detector_list += [detector.AsMarkdown for detector in self.Detectors[detect_kind].values()]
        detector_list = detector_list if len(detector_list) > 0 else ["None"]
        # Set up list of features
        feature_summary = ["## Processed Features",
                           "The features/metrics calculated from this game's event logs by OpenGameData when an 'export' is run."
                          ]
        feature_list = [feature.AsMarkdown for feature in self._aggregate_feats.values()] + [feature.AsMarkdown for feature in self._percount_feats.values()]
        feature_list = feature_list if len(feature_list) > 0 else ["None"]
        # Include other elements
        other_summary = ["## Other Elements",
                         "Other (potentially non-standard) elements specified in the game's schema, which may be referenced by event/feature processors."
                         ]
        other_element_list = [ f"{key} : {self._other_elements[key]}" for key in self._other_elements.keys()]
        other_range_summary = ["### Other Ranges",
                         "Extra ranges specified in the game's schema, which may be referenced by event/feature processors."
                         ]
        other_range_list = [ f"{key} : {self.OtherRanges[key]}" for key in self.OtherRanges ]

        ret_val = "  \n\n".join(event_summary
                              + enum_list + game_state_list + user_data_list + event_list
                              + detector_summary + detector_list
                              + feature_summary + feature_list
                              + other_summary + other_element_list
                              + other_range_summary + other_range_list)

        return ret_val


    # *** PUBLIC STATICS ***
    @staticmethod
    def FromFile(game_id:str, schema_path:Optional[Path] = None):
        # Give schema_path a default, don't think we can use game_id to construct it directly in the function header (so do it here if None)
        schema_path = schema_path or Path("./") / "ogd" / "games" / game_id / "schemas"
        all_elements = GameSchema._loadSchemaFile(game_name=game_id, schema_path=schema_path)
        return GameSchema(name=game_id, all_elements=all_elements or {})

    # *** PUBLIC METHODS ***

    def GetCountRange(self, count:Any) -> range:
        """Function to get a predefined range for per-count features, or to generate a range up to given count.
        Typically, this would be used to retrieve the `level_range` for the game.
        However, any other ranges defined in the game's schema can be retrieved here, or a custom range object generated (using `int(count)`).

        :param count: The name of a range defined in the game schema, or an object that can be int-ified to define a custom range.
        :type count: Any
        :return: The range object with name given by `count`, or a new range from 0 to (but not including) `int(count)`
        :rtype: range
        """
        if isinstance(count, str):
            if count.lower() == "level_range":
                count_range = self.LevelRange
            elif count in self.OtherRanges.keys():
                count_range = self.OtherRanges.get(count, range(0))
            else:
                other_range : Dict[str, int] = self.NonStandardElements.get(count, {'min':0, 'max':-1})
                count_range = range(other_range['min'], other_range['max']+1)
        else:
            count_range = range(0,int(count))
        return count_range

    def DetectorEnabled(self, detector_name:str, iter_mode:IterationMode, extract_mode:ExtractionMode) -> bool:
        """Function to check if detector with given base name and iteration mode (aggregate or percount) is enabled for given extract mode.

        :param detector_name: The base name of the detector class to check
        :type detector_name: str
        :param iter_mode: The "iteration" mode of the detector class (aggregate or per-count)
        :type iter_mode: IterationMode
        :param extract_mode: The extraction mode of the detector (which... should always be detector?)
        :type extract_mode: ExtractionMode
        :raises ValueError: Error indicating an unrecognized iteration mode was given.
        :return: True if the given detector is enabled in the schema, otherwise False
        :rtype: bool
        """
        if self._legacy_mode:
            return False
        ret_val : bool

        _detector_schema : Optional[DetectorSchema]
        match iter_mode:
            case IterationMode.AGGREGATE:
                _detector_schema = self.Detectors['aggregate'].get(detector_name)
            case IterationMode.PERCOUNT:
                _detector_schema = self.Detectors['per_count'].get(detector_name, self.Detectors['perlevel'].get(detector_name))
            case _:
                raise ValueError(f"In GameSchema, DetectorEnabled was given an unrecognized iteration mode of {iter_mode.name}")
        if _detector_schema is not None:
            ret_val = extract_mode in _detector_schema.Enabled
        else:
            Logger.Log(f"Could not find detector {detector_name} in schema for {iter_mode.name} mode")
            ret_val = False
        return ret_val

    def FeatureEnabled(self, feature_name:str, iter_mode:IterationMode, extract_mode:ExtractionMode) -> bool:
        if self._legacy_mode:
            return feature_name == "legacy"
        ret_val : bool

        _feature_schema : Optional[FeatureSchema]
        match iter_mode:
            case IterationMode.AGGREGATE:
                _feature_schema = self.AggregateFeatures.get(feature_name)
            case IterationMode.PERCOUNT:
                _feature_schema = self.PerCountFeatures.get(feature_name)
            case _:
                raise ValueError(f"In GameSchema, FeatureEnabled was given an unrecognized iteration mode of {iter_mode.name}")
        if _feature_schema is not None:
            ret_val = extract_mode in _feature_schema.Enabled
        else:
            Logger.Log(f"Could not find feature {feature_name} in schema for {iter_mode.name} mode")
            ret_val = False
        return ret_val

    def EnabledDetectors(self, iter_modes:Set[IterationMode], extract_modes:Set[ExtractionMode]=set()) -> Dict[str, DetectorSchema]:
        if self._legacy_mode:
            return {}
        ret_val : Dict[str, DetectorSchema] = {}

        if IterationMode.AGGREGATE in iter_modes:
            ret_val.update({key:val for key,val in self.AggregateDetectors.items() if val.Enabled.issuperset(extract_modes)})
        if IterationMode.PERCOUNT in iter_modes:
            ret_val.update({key:val for key,val in self.PerCountDetectors.items() if val.Enabled.issuperset(extract_modes)})
        return ret_val

    def EnabledFeatures(self, iter_modes:Set[IterationMode]={IterationMode.AGGREGATE, IterationMode.PERCOUNT}, extract_modes:Set[ExtractionMode]=set()) -> Dict[str, FeatureSchema]:
        if self._legacy_mode:
            return {"legacy" : AggregateSchema("legacy", {"type":"legacy", "return_type":None, "description":"", "enabled":True})} if IterationMode.AGGREGATE in iter_modes else {}
        ret_val : Dict[str, FeatureSchema] = {}

        if IterationMode.AGGREGATE in iter_modes:
            ret_val.update({key:val for key,val in self.AggregateFeatures.items() if val.Enabled.issuperset(extract_modes)})
        if IterationMode.PERCOUNT in iter_modes:
            ret_val.update({key:val for key,val in self.PerCountFeatures.items() if val.Enabled.issuperset(extract_modes)})
        return ret_val

    # *** PRIVATE STATICS ***

    @staticmethod
    def _loadSchemaFile(game_name:str, schema_path:Path) -> Optional[Dict[Any, Any]]:
        ret_val = None

        # 1. make sure the name and path are in the right form.
        schema_name = f"{game_name.upper()}.json"
        # 2. try to actually load the contents of the file.
        try:
            ret_val = utils.loadJSONFile(filename=schema_name, path=schema_path)
        except (ModuleNotFoundError, FileNotFoundError) as err:
            Logger.Log(f"Unable to load GameSchema for {game_name}, {schema_name} does not exist! Trying to load from json template instead...", logging.WARN, depth=1)
            ret_val = GameSchema._schemaFromTemplate(schema_path=schema_path, schema_name=schema_name)
            if ret_val is not None:
                Logger.Log(f"Loaded schema for {game_name} from template.", logging.WARN, depth=1)
            else:
                Logger.Log(f"Failed to load schema for {game_name} from template.", logging.WARN, depth=1)
        else:
            if ret_val is None:
                Logger.Log(f"Could not load game schema at {schema_path / schema_name}, the file was empty!", logging.ERROR)
        return ret_val

    @staticmethod
    def _schemaFromTemplate(schema_path:Path, schema_name:str) -> Optional[Dict[Any, Any]]:
        ret_val = None

        template_name = schema_name + ".template"
        try:
            ret_val = loadJSONFile(filename=template_name, path=schema_path, autocorrect_extension=False)
        except FileNotFoundError as no_file:
            Logger.Log(       f"Could not load {schema_name} from template, the template does not exist at {schema_path}.", logging.WARN, depth=2)
            print(f"(via print) Could not create {schema_name} from template, the template does not exist at {schema_path}.")
        else:
            Logger.Log(f"Trying to copy {schema_name} from template, for future use...", logging.DEBUG, depth=2)
            template = schema_path / template_name
            try:
                copyfile(template, schema_path / schema_name)
            except Exception as cp_err:
                Logger.Log(       f"Could not copy {schema_name} from template, a {type(cp_err)} error occurred:\n{cp_err}", logging.WARN, depth=2)
                print(f"(via print) Could not copy {schema_name} from template, a {type(cp_err)} error occurred:\n{cp_err}")
            else:
                Logger.Log(       f"Successfully copied {schema_name} from template.", logging.DEBUG, depth=2)
        return ret_val


    @staticmethod
    def _parseEnumDefs(enums_list:Dict[str, Any]) -> Dict[str, List[str]]:
        ret_val : Dict[str, List[str]]
        if isinstance(enums_list, dict):
            ret_val = enums_list
        else:
            ret_val = {}
            Logger.Log(f"enums_list was unexpected type {type(enums_list)}, defaulting to empty Dict.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseGameState(game_state:Dict[str, Any]) -> Dict[str, DataElementSchema]:
        ret_val : Dict[str, DataElementSchema]
        if isinstance(game_state, dict):
            ret_val = {name:DataElementSchema(name=name, all_elements=elems) for name,elems in game_state.items()}
        else:
            ret_val = {}
            Logger.Log(f"game_state was unexpected type {type(game_state)}, defaulting to empty dict.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseUserData(user_data:Dict[str, Any]) -> Dict[str, DataElementSchema]:
        ret_val : Dict[str, DataElementSchema]
        if isinstance(user_data, dict):
            ret_val = {name:DataElementSchema(name=name, all_elements=elems) for name,elems in user_data.items()}
        else:
            ret_val = {}
            Logger.Log(f"user_data was unexpected type {type(user_data)}, defaulting to empty dict.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseEventList(events_list:Dict[str, Any]) -> List[EventSchema]:
        ret_val : List[EventSchema]
        if isinstance(events_list, dict):
            ret_val = [EventSchema(name=key, all_elements=val) for key,val in events_list.items()]
        else:
            ret_val = []
            Logger.Log(f"events_list was unexpected type {type(events_list)}, defaulting to empty List.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseDetectorMap(detector_map:Dict[str, Any]) -> DetectorMapSchema:
        ret_val : DetectorMapSchema
        if isinstance(detector_map, dict):
            ret_val = DetectorMapSchema(name=f"Detectors", all_elements=detector_map)
        else:
            ret_val = DetectorMapSchema(name="Empty Features", all_elements={})
            Logger.Log(f"detector_map was unexpected type {type(detector_map)}, defaulting to empty map.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseFeatureMap(feature_map:Dict[str, Any]) -> FeatureMapSchema:
        ret_val : FeatureMapSchema
        if isinstance(feature_map, dict):
            ret_val = FeatureMapSchema(name=f"Features", all_elements=feature_map)
        else:
            ret_val = FeatureMapSchema(name="Empty Features", all_elements={})
            Logger.Log(f"feature_map was unexpected type {type(feature_map)}, defaulting to empty map.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseLevelRange(level_range:Dict[str, int]) -> Tuple[Optional[int], Optional[int]]:
        ret_val : Tuple[Optional[int], Optional[int]]
        if isinstance(level_range, dict):
            ret_val = (level_range.get("min", None), level_range.get("max", None))
        else:
            ret_val = (None, None)
            Logger.Log(f"level_range was unexpected type {type(level_range)}, defaulting to no specified range.", logging.WARN)
        return ret_val

    # *** PRIVATE METHODS ***



"""Initializer for ThermoVR features"""

__all__ = [
    "LabCompleteCount",
    "LeftHandMoves",
    "RightHandMoves",
    "PhasesReached",
    "PlayMode",
    "TaskCompleteCount",
    "ToolNudgeCount",
    "ToolSliderTime",

]

from . import LabCompleteCount
from . import LeftHandMoves
from . import RightHandMoves
from . import PhasesReached
from . import PlayMode
from . import TaskCompleteCount
from . import ToolNudgeCount
from . import ToolSliderTime



import abc
import typing
import logging
from abc import ABC
from statistics import median
import pandas as pd
import numpy as np
import json
from datetime import datetime as dt
## import local libraries
from models.SequenceModel import SequenceModel
from models.DeathPredModel import DeathPredModel

EMOTE_DESPERATE_ENUM = 2
FB_DEATH_ENUM = 18
EMOTE_EVT_ENUM = 24
MONEY_RATE_ENUM = 33
FOOD_AV_ENUM = 32
ACHIEV_ENUM = 17

class DeathThresholdModel(DeathPredModel):
    def __init__(self, levels=[]):
        super().__init__()
        self.thresh_df = pd.DataFrame()
        self.res_df = pd.DataFrame()

    def Eval(self, rows):
        return self._eval(rows)

    def _eval(self, rows):
        self.df = pd.DataFrame.from_records(rows)
        self.process_df()
        return self.matcher()

    def matcher(self):
        self.process_thresh_df()
        i_df = self.link_deaths_yums()
        i_df = i_df.merge(
            self.get_money_food_data(i_df), right_on='key', left_on=i_df.index, how='left')
        self.output_thresh_data(i_df)
        return self.res_df

    def process_thresh_df(self):
        self.filtered_df['key'] = self.filtered_df[['sess_id', 'name', 'event_custom', 'emote_enum']].apply(
            lambda x: self.f(*x), axis=1
        )
        self.filtered_df['fb_n'] = self.filtered_df['sess_id'].apply(
            lambda x: self.f2(x)
        )
        self.filtered_df['unique_keys_n'] = self.filtered_df[['sess_id', 'key']].apply(
            lambda x: self.f3(*x), axis=1
        )
        self.thresh_df = self.filtered_df.sort_values(
            by=['sess_id', 'utc_time_secs', 'key', 'event_custom'], ascending=[True, True, True, False])

        self.thresh_df = self.thresh_df.drop(
            list(self.thresh_df[self.thresh_df['unique_keys_n'] < 2].index), axis=0)
        # drop rows with only one key event

    def link_deaths_yums(self):
        t_df, i_df = self.thresh_df, pd.DataFrame()
        d_list = [x for x in t_df.itertuples() if x.event_custom == 24 and x.emote_enum == 2]
        for i in range(len(d_list)):
            # & (t_df['key']!=x.key)
            x = d_list[i]
            s = t_df[
                (t_df['sess_id'] == x.sess_id) &
                (t_df['utc_time_secs'] > x.utc_time_secs) &
                (t_df['name'] == x.name) &
                (t_df['key'] != x.key)].sort_values(by=['utc_time_secs'])
            el_dict = {
                "sess_id": x.sess_id, "h_key": x.key, "h_time": x.utc_time_secs, "fb": x.name, "fb_n": x.fb_n}
            if len(s.index) > 0:
                el = s.iloc[0, :]
                if el.key == x.key:
                    el = s.iloc[1, :]

                if el.event_custom == 24:
                    if el.emote_enum == 7:
                        el_dict['y_key'], el_dict['y_time'] = el.key, el.utc_time_secs
                        i_df = i_df.append(el_dict, ignore_index=True)
                    else:
                        raise Exception('code not filtering out desperates')
                elif el.event_custom == 18:
                    el_dict['d_key'], el_dict['d_time'] = el.key, el.utc_time_secs
                    i_df = i_df.append(el_dict, ignore_index=True)
                else:
                    raise Exception('random event detected')

        i_df = i_df.fillna(value={"y_key": "None", "d_key": "None", "y_time": -1, "d_time": -1})
        return i_df

    def get_money_food_data(self, i_df):
        food_money_df = pd.DataFrame()
        for x in i_df.itertuples():
            query = f'((sess_id == {x.sess_id}) & (event_custom == {FOOD_AV_ENUM}) | (event_custom == {MONEY_RATE_ENUM}))'
            if x.d_key == "None":
                query += f' & ((utc_time_secs > {x.h_time}) & (utc_time_secs < {x.y_time}))'
            elif x.y_key == "None":
                query += f' & ((utc_time_secs > {x.h_time}) & (utc_time_secs < {x.d_time}))'
            else:
                raise Exception("BOTH NONE?")
            s = self.df.query(query)
            fm_df = pd.DataFrame(
                list(s['event_data_complex'].values))
            fm_df['key'] = x.Index
            fm_df = fm_df.fillna(0)
            fm_df = fm_df.drop_duplicates()
            food_money_df = food_money_df.append(fm_df, ignore_index=True, sort=False)
        return self.money_food_thresh_data(food_money_df, i_df)

    def money_food_thresh_data(self, food_money_df, i_df):
        thresh_f = pd.DataFrame()
        for i, x in enumerate(i_df.itertuples()):
            fm_df, curr_fb = food_money_df.query(f'(key) == {i}'), 0
            money = sum([y.rate for y in fm_df.itertuples() if y.event_custom == 33])
            food = sum([y.food for y in fm_df.itertuples() if y.event_custom == 32])
            v_time = x.y_time if x.y_key != "None" else x.d_time
            gs_df = self.df.query(f'(event_custom == {0}) & (utc_time_secs < {v_time}) & (utc_time_secs > {x.h_time})')
            if len(gs_df.index) > 0:
                curr_fb_list = list(set(gs_df['event_data_complex'].apply(lambda x: self.f5(x), axis=1)))
                if len(curr_fb_list) == 1:
                    if curr_fb_list[0] < 1:
                        curr_fb = x.fb_n
                    else:
                        curr_fb = curr_fb_list[0]
                elif max(curr_fb_list) < 1:
                    curr_fb = x.fb_n
                else:
                    curr_fb = max(curr_fb_list)
            thresh_f = thresh_f.append({
                "key": x.Index,
                "avg_money": money / (curr_fb or 1),
                "avg_food": food / (curr_fb or 1),
                "total_money": money,
                "total_food": food,
                "interval_secs": v_time - x.h_time,
                "curr_fb_n": curr_fb
            }, ignore_index=True)

        return thresh_f

    def output_thresh_data(self, i_df):
        th_df = pd.DataFrame()
        for x in list(i_df['curr_fb_n'].drop_duplicates()):
            food_y = [r.total_food for r in i_df[i_df['curr_fb_n'] == x].itertuples() if r.y_key != "None"] + [0]
            food_d = [r.total_food for r in i_df[i_df['curr_fb_n'] == x].itertuples() if r.d_key != "None"] + [0]
            money_y = [r.total_money for r in i_df[i_df['curr_fb_n'] == x].itertuples() if r.y_key != "None"] + [0]
            money_d = [r.total_money for r in i_df[i_df['curr_fb_n'] == x].itertuples() if r.d_key != "None"] + [0]
            interval_y = [r.interval_secs for r in i_df[i_df['curr_fb_n'] == x].itertuples() if r.y_key != "None"] + [0]
            interval_d = [r.interval_secs for r in i_df[i_df['curr_fb_n'] == x].itertuples() if r.d_key != "None"] + [0]
            th_df = th_df.append(
                {
                    "fb_num": x, "max_food_to_yum": max(food_y),
                    "min_food_to_yum": min(food_y), "max_food_to_death": max(food_d),
                    "min_food_to_death": min(food_d), "max_money_to_yum": max(money_y),
                    "min_money_to_yum": min(money_y), "max_money_to_death": max(money_d),
                    "min_money_to_death": min(money_d), "avg_food_to_yum": sum(food_y) / len(food_y),
                    "avg_food_to_death": sum(food_d) / len(food_d), "avg_money_to_yum": sum(money_y) / len(money_y),
                    "avg_money_to_death": sum(money_d) / len(money_d),
                    "avg_interval_to_death": sum(interval_d) / len(interval_d),
                    "max_interval_to_death": max(interval_d), "min_interval_to_death": min(interval_d),
                    "avg_interval_to_yum": sum(interval_y) / len(interval_y), "max_interval_to_yum": max(interval_y),
                    "min_interval_to_yum": min(interval_y), "median_food_to_yum": median(food_y),
                    "median_food_to_death": median(food_d), "median_money_to_death": median(money_d),
                    "median_money_to_yum": median(money_y),
                }, ignore_index=True)
        self.res_df = th_df.sort_values(by=['fb_num']).set_index('fb_num')

    def f(self, s, n, v, e):
        if v == 24:
            return f'{int(n)}_{int(v)}_{int(e)}'
        else:
            return f'{int(n)}_{int(v)}'

    def f2(self, x):
        return len(
            self.filtered_df[self.filtered_df['sess_id'] == x]['name'].drop_duplicates())

    def f3(self, x, y):
        return len(
            self.filtered_df[(self.filtered_df['sess_id'] == x)]['key'].drop_duplicates().index
        )

    def read_stringified_array(self, arr: str):
        if not arr:
            return []
        return [int(x) for x in arr.split(',')]

    def f5(self, x):
        return len(self.read_stringified_array(x['farmbits'])) / 9


## import standard libraries
import abc
import logging
from typing import List, Optional
# import locals
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.utils.Logger import Logger

## @class ExtractorParams
class GeneratorParameters:
    """Dumb struct to hold the data that should be available to every Generator.
    This just makes it easier to add/manage any new params,
    so that we don't need to change the param list for hundreds of individual
    extractor subclasses every time something changes.
    """
    def __init__(self, name:str, description:str, mode:ExtractionMode, count_index:Optional[int]):
        self._name = name
        self._desc = description
        self._mode = mode
        self._count_index = count_index

## @class Extractor
#  Abstract base class for all data extractors (features and detectors)
class Generator(abc.ABC):
#TODO: use a dirty bit so we only run the GetValue function if we've received an event or feature since last calculation

    # *** ABSTRACTS ***

    ## Abstract function to get a list of event types the Feature wants.
    @classmethod
    @abc.abstractmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """ Abstract function to get a list of event types the Feature wants.
            The types of event accepted by a feature are a responsibility of the Feature's developer,
            so this is a required part of interface instead of a config item in the schema.

        :return: [description]
        :rtype: List[str]
        """
        raise TypeError(f"Can't call function on class {cls.__name__} with abstract method _eventFilter")

    @classmethod
    @abc.abstractmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        """Base function for getting any features a second-order feature depends upon.
        By default, no dependencies.
        Any feature intented to be second-order should override this function.

        :return: _description_
        :rtype: List[str]
        """
        raise TypeError(f"Can't call function on class {cls.__name__} with abstract method _featureFilter")

    ## Abstract declaration of a function to perform update of a feature from a row.
    @abc.abstractmethod
    def _updateFromEvent(self, event:Event):
        """Abstract declaration of a function to perform update of a feature from a row.

        :param event: An event, used to update the feature's data.
        :type event: Event
        """
        pass

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, params:GeneratorParameters):
        self._params = params

    def __str__(self):
        return f"{self.Name} : {self.Description}"

    # *** PUBLIC STATICS ***

    @staticmethod
    def WarningMessage(message:str):
        """
        Simple wrapper for the Logger in utils, makes it easier for subclasses to print warning messages without extra imports.

        :param message: The warning message to print
        :type message: str
        """
        Logger.Log(message=message, level=logging.WARN, depth=3)

    ## Base function to get the minimum log version the feature can handle.
    @staticmethod
    def MinVersion() -> Optional[str]:
        """ Base function to get the minimum log version the feature can handle.
            A value of None will set no minimum, so all levels are accepted (unless a max is set).
            Typically default to None, unless there is a required element of the event data that was not added until a certain version.        
            The versions of data accepted by a feature are a responsibility of the Feature's developer,
            so this is a required part of interface instead of a config item in the schema.

        :return: [description]
        :rtype: Optional[str]
        """
        return None

    ## Base function to get the maximum log version the feature can handle.
    @staticmethod
    def MaxVersion() -> Optional[str]:
        """ Base function to get the maximum log version the feature can handle.
            A value of None will set no maximum, so all levels are accepted (unless a min is set).
            Typically default to None, unless the feature is not compatible with new data and is only kept for legacy purposes.
            The versions of data accepted by a feature are a responsibility of the Feature's developer,
            so this is a required part of interface instead of a config item in the schema.

        :return: [description]
        :rtype: Optional[str]
        """
        return None

    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        """List of ExtractionMode supported by the Extractor

        Base function to give a list of which ExtractionModes an extractor will handle.
        :return: _description_
        :rtype: List[ExtractionMode]
        """
        return [ExtractionMode.POPULATION, ExtractionMode.PLAYER, ExtractionMode.SESSION, ExtractionMode.DETECTOR]

    # *** PUBLIC METHODS ***

    @classmethod
    def EventFilter(cls, mode:ExtractionMode) -> List[str]:
        return cls._eventFilter(mode=mode)

    @classmethod
    def FeatureFilter(cls, mode:ExtractionMode) -> List[str]:
        return cls._featureFilter(mode=mode)

    def UpdateFromEvent(self, event:Event):
        if self._validateEvent(event=event):
            self._updateFromEvent(event=event)

    # *** PROPERTIES ***

    @property
    def Name(self) -> str:
        return self._params._name

    @property
    def Description(self) -> str:
        return self._params._desc

    @property
    def ExtractionMode(self) -> ExtractionMode:
        return self._params._mode

    @property
    def CountIndex(self) -> Optional[int]:
        return self._params._count_index

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***

    def _validateEvent(self, event:Event) -> bool:
        """Private function to check if a given event has valid version and type for this Feature.

        :param event: The event to be checked.
        :type event: Event
        :return: True if the event has valid version and type, otherwise false.
        :rtype: bool
        """
        return (
            self._validateVersion(event.LogVersion)
        and self._validateEventType(event_type=event.EventName)
        )

    ## Private function to check whether the given data version from a row is acceptable by this feature extractor.
    def _validateVersion(self, data_version:str) -> bool:
        """Private function to check whether a given version is valid for this Feature.

        :param data_version: The logging version for some event to be checked.
        :type data_version: str
        :return: True if the given version is valid for this feature, otherwise false.
        :rtype: bool
        """
        if data_version != 'None':
            min = self.MinVersion()
            if min is not None:
                if Event.CompareVersions(data_version, min) < 0:
                    return False # too old, not valid.
            max = self.MaxVersion()
            if max is not None:
                if Event.CompareVersions(data_version, max) > 0:
                    return False # too new, not valid
            return True # passed both cases, valid.
        else:
            return False # data_version of None is invalid.

    def _validateEventType(self, event_type:str) -> bool:
        """Private function to check whether a given event type is accepted by this Feature.

        :param event_type: The name of the event type to be checked.
        :type event_type: str
        :return: True if the given event type is in this feature's list, otherwise false.
        :rtype: bool
        """
        _deps = self.EventFilter(mode=self.ExtractionMode)
        if event_type in _deps or 'all_events' in _deps:
            return True
        else:
            return False


# import standard libraries
import logging
import traceback
from typing import List, Dict, Type, Optional, Set
# import local files
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.registries.ExtractorRegistry import ExtractorRegistry
from ogd.core.processors.ExtractorProcessor import ExtractorProcessor
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExportMode import ExportMode
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.utils.Logger import Logger
from ogd.core.utils.utils import ExportRow

## @class SessionProcessor
#  Class to extract and manage features for a processed csv file.
class SessionProcessor(ExtractorProcessor):

    # *** BUILT-INS & PROPERTIES ***

    ## Constructor for the SessionProcessor class.
    def __init__(self, LoaderClass:Type[GeneratorLoader], game_schema: GameSchema, player_id:str, session_id:str,
                 feature_overrides:Optional[List[str]]=None):
        """Constructor for the SessionProcessor class.
        Simply stores some data for use later, including the type of extractor to
        use.

        :param LoaderClass: The type of data extractor to use for input data.
                            This should correspond to whatever game_id is in the TableSchema.
        :type LoaderClass: Type[GeneratorLoader]
        :param game_schema: A dictionary that defines how the game data itself is structured.
        :type game_schema: GameSchema
        :param player_id: _description_
        :type player_id: str
        :param session_id: _description_
        :type session_id: str
        :param feature_overrides: _description_, defaults to None
        :type feature_overrides: Optional[List[str]], optional
        :param session_file: _description_, defaults to None
        :type session_file: Union[IO[str],None], optional
        """
        ## Define instance vars
        self._session_id   : str = session_id
        self._player_id    : str = player_id
        # NOTE: need session and player IDs set before we do initialization in parent.
        super().__init__(LoaderClass=LoaderClass, game_schema=game_schema, feature_overrides=feature_overrides)

    def __str__(self):
        return f"SessionProcessor({self._player_id}, {self._session_id})"

    def __repr__(self):
        return str(self)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @property
    def _mode(self) -> ExtractionMode:
        return ExtractionMode.SESSION

    @property
    def _playerID(self) -> str:
        return self._player_id

    @property
    def _sessionID(self) -> str:
        return self._session_id

    def _getGeneratorNames(self) -> List[str]:
        return ["PlayerID", "SessionID"] + self._registry.GetGeneratorNames()

    ## Function to handle processing of a single row of data.
    def _processEvent(self, event: Event):
        """Function to handle processing of a single row of data.
        Basically just responsible for ensuring an extractor for the session corresponding
        to the row already exists, then delegating the processing to that extractor.

        :param event: Object with the data for the game event data to be processed.
        :type event: Event
        """
        self._registry.UpdateFromEvent(event)

    def _getLines(self) -> List[ExportRow]:
        ret_val : ExportRow
        # if as_str:
        #     ret_val = [self._playerID, self._sessionID] + self._registry.GetFeatureStringValues()
        # else:
        ret_val = [self._playerID, self._sessionID] + self._registry.GetFeatureValues()
        return [ret_val]

    def _getFeatureData(self, order:int) -> List[FeatureData]:
        return self._registry.GetFeatureData(order=order, player_id=self._player_id, sess_id=self._session_id)

    def _clearLines(self) -> None:
        Logger.Log(f"Clearing features from SessionProcessor for player {self._player_id}, session {self._session_id}.", logging.DEBUG, depth=2)
        self._registry = ExtractorRegistry(mode=self._mode)

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



from typing import Any, List, Set
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
import json

class TaskCompleteCount(SessionFeature):

    def __init__(self, params: GeneratorParameters):
        self.completed_tasks: Set[str] = set()
        super().__init__(params=params)

    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["target_state_achieved", "click_submit_answer"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event: Event) -> None:
        if event.EventName == "target_state_achieved":
            target_state = event.EventData.get("target_state")
            if target_state:
                self.completed_tasks.add(str(target_state))
        elif event.EventName == "click_submit_answer":
            quiz_task = event.EventData.get("quiz_task", [])
            if isinstance(quiz_task, str):
                quiz_task = json.loads(quiz_task)
            if quiz_task and quiz_task.get("is_complete"):
                _id = f"{quiz_task.get('lab_name')}_{quiz_task.get('section_number')}_{quiz_task.get('task_number')}"
                self.completed_tasks.add(_id)

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [len(self.completed_tasks)]



"""Initializer for Penguins features"""

__all__ = [
    "ActivityCompleted",
    "ActivityDuration",
    "BuiltNestCount",
    "BuiltWrongNestCount",
    "EatFishCount",
    "EggLostCount",
    "EggRecoverTime",
    "GazeCount",
    "GazeDuration",
    "LogVersion",
    "MirrorWaddleDuration",
    "PenguinInteractCount",
    "PlayerInactiveAvgDuration",
    "RegionDuration",
    "RegionEnterCount",
    "RegionsEncountered",
    "RingChimesCount",
    "RockBashCount",
    "RockMultiplePickupCount",
    "RockPickupCount",
    "SessionDuration",
    "SkuaBashCount",
    "SkuaPeckCount",
    "SnowBallDuration",
    "WaddleCount",
    "WaddlePerRegion",
    "WaddlePerRegion"
]
# aggregated features

from . import BuiltNestCount
from . import BuiltWrongNestCount
from . import EatFishCount
from . import EggLostCount
from . import EggRecoverTime
from . import GazeCount
from . import GazeDuration
from . import LogVersion
from . import MirrorWaddleDuration
from . import PenguinInteractCount
from . import PlayerInactiveAvgDuration
from . import RegionsEncountered
from . import RingChimesCount
from . import RockBashCount
from . import RockMultiplePickupCount
from . import RockPickupCount
from . import SessionDuration
from . import SkuaBashCount
from . import SkuaPeckCount
from . import SnowBallDuration
from . import WaddleCount

# percount features
from . import ActivityCompleted
from . import ActivityDuration
from . import RegionDuration
from . import RegionEnterCount
from . import WaddlePerRegion



# import libraries
from typing import Any, List, Optional
from ogd.core.generators.Generator import GeneratorParameters
# import local files
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.models.Event import Event

class UserEnabled(SessionFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._full_screen : Optional[int] = None
        self._music : Optional[int] = None
        self._hq : Optional[int] = None

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["CUSTOM.1"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return [] 

    def _updateFromEvent(self, event:Event) -> None:
        _data = event.EventData
        true_values = ["1", "true"]
        for item in ["fullscreen", "music", "hq"]:
            if _data.get(item) is None:
                raise(ValueError(f"Can't find {item} item in the event data!"))
        if str(_data.get("fullscreen")).lower() in true_values:
            self._full_screen = 1
        else:
            self._full_screen = 0

        if str(_data.get("music")).lower() in true_values:
            self._music = 1
        else:
            self._music = 0

        if str(_data.get("hq")).lower() in true_values:
            self._hq = 1
        else:
            self._hq = 0

        return

    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._full_screen, self._music, self._hq]

    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["Music", "HQGraphics"] 

    def BaseFeatureSuffix(self) -> str:
        return "FullScreen"


import json
import logging
import os
from datetime import datetime, date
from google.cloud import bigquery
from google.api_core.exceptions import BadRequest
from typing import Dict, Final, List, Tuple, Optional
# import locals
from ogd.core.interfaces.EventInterface import EventInterface
from ogd.core.models.enums.IDMode import IDMode
from ogd.core.schemas.configs.GameSourceSchema import GameSourceSchema
from ogd.core.schemas.configs.data_sources.BigQuerySourceSchema import BigQuerySchema
from ogd.core.utils.Logger import Logger

AQUALAB_MIN_VERSION : Final[float] = 6.2

class BigQueryInterface(EventInterface):

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, game_id:str, config:GameSourceSchema, fail_fast:bool):
        super().__init__(game_id=game_id, config=config, fail_fast=fail_fast)
        self.Open()

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    def _open(self, force_reopen: bool = False) -> bool:
        if force_reopen:
            self.Close()
            self.Open(force_reopen=False)
        if not self._is_open:
            if "GITHUB_ACTIONS" in os.environ:
                self._client = bigquery.Client()
            elif isinstance(self._config.Source, BigQuerySchema):
                os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = self._config.Source.Credential or "NO CREDENTIAL CONFIGURED!" or f"./{self._game_id}.json"
                self._client = bigquery.Client()
            else:
                raise ValueError("No BigQuery credential available in current configuration!")
            if self._client != None:
                self._is_open = True
                Logger.Log("Connected to BigQuery database.", logging.DEBUG)
                return True
            else:
                Logger.Log("Could not connect to BigQuery Database.", logging.WARN)
                return False
        else:
            return True

    def _close(self) -> bool:
        self._client.close()
        self._is_open = False
        Logger.Log("Closed connection to BigQuery.", logging.DEBUG)
        return True

    def _allIDs(self) -> List[str]:
        ret_val = []

        query = f"""
            SELECT DISTINCT session_id
            FROM `{self.DBPath()}`,
        """
        Logger.Log(f"Running query for all ids:\n{query}", logging.DEBUG, depth=3)
        try:
            data = self._client.query(query)
            session_ids = [str(row['session_id']) for row in data]
        except BadRequest as err:
            Logger.Log(f"In _allIDs, got a BadRequest error when trying to retrieve data from BigQuery, defaulting to empty result!\n{err}")
        else:
            ret_val = session_ids
        return ret_val

    def _fullDateRange(self) -> Dict[str, datetime]:
        ret_val : Dict[str, datetime] = {}

        query = f"""
            SELECT MIN(server_time), MAX(server_time)
            FROM `{self.DBPath()}`
        """
        Logger.Log(f"Running query for full date range:\n{query}", logging.DEBUG, depth=3)
        try:
            data = list(self._client.query(query))
            date_range : Dict[str, datetime] = { 'min':data[0][0], 'max':data[0][1] }
        except BadRequest as err:
            Logger.Log(f"In _fullDateRange, got a BadRequest error when trying to retrieve data from BigQuery, defaulting to empty result!\n{err}")
        else:
            ret_val = date_range
        return ret_val

    def _rowsFromIDs(self, id_list:List[str], id_mode:IDMode=IDMode.SESSION, versions:Optional[List[int]] = None, exclude_rows:Optional[List[str]]=None) -> List[Tuple]:
        # 2) Set up clauses to select based on Session ID or Player ID.
        ret_val = []
        if self._client != None:
            query = self._generateRowFromIDQuery(id_list=id_list, id_mode=id_mode, exclude_rows=exclude_rows)
            Logger.Log(f"Running query for rows from IDs:\n{query}", logging.DEBUG, depth=3)
            try:
                data = self._client.query(query)
                Logger.Log(f"...Query yielded results, with query in state: {data.state}", logging.DEBUG, depth=3)
            except BadRequest as err:
                Logger.Log(f"In _rowsFromIDs, got a BadRequest error when trying to retrieve data from BigQuery, defaulting to empty result!\n{err}")
            else:
                for row in data:
                    items = tuple(row.items())
                    event = []
                    for item in items:
                        match item[0]:
                            case "event_params":
                                _params = {param['key']:param['value'] for param in item[1]}
                                event.append(json.dumps(_params, sort_keys=True))
                            case "device":
                                event.append(json.dumps(item[1], sort_keys=True))
                            case _:
                                event.append(item[1])
                    ret_val.append(tuple(event))
        return ret_val

    def _IDsFromDates(self, min:datetime, max:datetime, versions:Optional[List[int]] = None) -> List[str]:
        ret_val = []
        str_min, str_max = min.strftime("%Y%m%d"), max.strftime("%Y%m%d")
        query = f"""
            SELECT DISTINCT session_id
            FROM `{self.DBPath(min_date=min.date(), max_date=max.date())}`
            WHERE _TABLE_SUFFIX BETWEEN '{str_min}' AND '{str_max}'
        """
        Logger.Log(f"Running query for ids from dates:\n{query}", logging.DEBUG, depth=3)
        try:
            data = self._client.query(query)
            ids = [str(row['session_id']) for row in data]
        except BadRequest as err:
            Logger.Log(f"In _IDsFromDates, got a BadRequest error when trying to retrieve data from BigQuery, defaulting to empty result!\n{err}")
        else:
            ret_val = ids
            Logger.Log(f"Found {len(ret_val)} ids. {ret_val if len(ret_val) <= 5 else ''}", logging.DEBUG, depth=3)
        return ret_val

    def _datesFromIDs(self, id_list:List[str], id_mode:IDMode=IDMode.SESSION, versions:Optional[List[int]] = None) -> Dict[str, datetime]:
        ret_val : Dict[str, datetime] = {}

        match id_mode:
            case IDMode.SESSION:
                id_string = ','.join([f"'{x}'" for x in id_list])
                where_clause = f"WHERE session_id IN ({id_string})"
            case IDMode.USER:
                id_string = ','.join([f"'{x}'" for x in id_list])
                where_clause = f"WHERE user_id IN ({id_string})"
            case _:
                Logger.Log(f"Invalid ID mode given (name={id_mode.name}, val={id_mode.value}), defaulting to session mode.", logging.WARNING, depth=3)
                id_string = ','.join([f"{x}" for x in id_list])
                where_clause = f"WHERE session_id IN ({id_string})"
        query = f"""
            SELECT MIN(server_time), MAX(server_time)
            FROM `{self.DBPath()}`
            {where_clause}
        """
        Logger.Log(f"Running query for dates from IDs:\n{query}", logging.DEBUG, depth=3)
        try:
            data = list(self._client.query(query))
            Logger.Log(f"...Query yielded results:\n{data}", logging.DEBUG, depth=3)
        except BadRequest as err:
            Logger.Log(f"In _datesFromIDs, got a BadRequest error when trying to retrieve data from BigQuery, defaulting to empty result!\n{err}", logging.WARNING)
        else:
            if len(data) == 1:
                dates = data[0]
                if len(dates) == 2 and dates[0] is not None and dates[1] is not None:
                    _min = dates[0] if type(dates[0]) == datetime else datetime.strptime(str(dates[0]), "%m-%d-%Y %H:%M:%S")
                    _max = dates[1] if type(dates[1]) == datetime else datetime.strptime(str(dates[1]), "%m-%d-%Y %H:%M:%S")
                    ret_val = {'min':_min, 'max':_max}
                else:
                    Logger.Log(f"BigQueryInterface query did not give both a min and a max, setting both to 'now'", logging.WARNING, depth=3)
                    ret_val = {'min':datetime.now(), 'max':datetime.now()}
            else:
                Logger.Log(f"BigQueryInterface query did not return any results, setting both min and max to 'now'", logging.WARNING, depth=3)
                ret_val = {'min':datetime.now(), 'max':datetime.now()}
        return ret_val

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    def IsOpen(self) -> bool:
        """Overridden version of IsOpen function, checks that BigQueryInterface client has been initialized.

        :return: True if the interface is open, else False
        :rtype: bool
        """
        return True if (super().IsOpen() and self._client is not None) else False

    def DBPath(self, min_date:Optional[date]=None, max_date:Optional[date]=None) -> str:
        """The path of form "[projectID].[datasetID].[tableName]" used to make queries

        :return: The full path from project ID to table name, if properly set in configuration, else the literal string "INVALID SOURCE SCHEMA".
        :rtype: str
        """
        if isinstance(self._config.Source, BigQuerySchema):
            # _current_date = datetime.now().date()
            date_wildcard = "*"
            # if min_date is not None and max_date is not None:
            #     date_wildcard = BigQueryInterface._datesWildcard(a=min_date, b=max_date)
            # elif min_date is not None:
            #     date_wildcard = BigQueryInterface._datesWildcard(a=min_date, b=_current_date)
            # elif max_date is not None:
            #     date_wildcard = BigQueryInterface._datesWildcard(a=_current_date, b=max_date)
            return f"{self._config.Source.AsConnectionInfo}.{self._config.DatabaseName}.{self._config.TableName}_{date_wildcard}"
        else:
            return "INVALID SOURCE SCHEMA"

    # *** PRIVATE STATICS ***

    @staticmethod
    def _datesWildcard(a:date, b:date) -> str:
        ret_val = "*"
        if a.year == b.year:
            ret_val = str(a.year).rjust(4, "0")
            if a.month == b.month:
                ret_val += str(a.month).rjust(2, "0")
                if a.day == b.day:
                    ret_val += str(a.day).rjust(2, "0")
                else:
                    ret_val += "*"
            else:
                ret_val += "*"
        else:
            ret_val = "*"
        return ret_val

    # *** PRIVATE METHODS ***

    def _generateRowFromIDQuery(self, id_list:List[str], id_mode:IDMode, exclude_rows:Optional[List[str]]=None) -> str:
        id_clause : str = ""
        id_string = ','.join([f"'{x}'" for x in id_list])
        match id_mode:
            case IDMode.SESSION:
                id_clause = f"session_id IN ({id_string})"
            case IDMode.USER:
                id_clause  = f"user_id IN ({id_string})"
            case _:
                Logger.Log(f"Invalid ID mode given (name={id_mode.name}, val={id_mode.value}), defaulting to session mode.", logging.WARNING, depth=3)
                id_clause = f"session_id IN ({id_string})"
        # 3) Set up WHERE clause based on whether we need Aqualab min version or not.
        where_clause = f" WHERE {id_clause}"
        if exclude_rows is not None:
            exclude_string = ','.join([f"'{x}'" for x in exclude_rows])
            where_clause += f" AND event_name not in ({exclude_string})"

        # 4) Set up actual query
        # TODO Order by user_id, and by timestamp within that.
        # Note that this could prove to be wonky when we have more games without user ids,
        # will need to really rethink this when we start using new system.
        # Still, not a huge deal because most of these will be rewritten at that time anyway.
        query = f"""
            SELECT *
            FROM `{self.DBPath()}`
            {where_clause}
            ORDER BY `user_id`, `session_id`, `event_sequence_index` ASC
        """
        return query



# import standard libraries
import logging
from pathlib import Path
from typing import Any, Dict, Optional
# import local files
from ogd.core.schemas.Schema import Schema
from ogd.core.utils.Logger import Logger

class FileIndexingSchema(Schema):
    def __init__(self, name:str, all_elements:Dict[str, Any]):
        self._local_dir     : Path
        self._remote_url    : Optional[str]
        self._templates_url : str

        if not isinstance(all_elements, dict):
            all_elements = {}
            Logger.Log(f"For {name} base config, all_elements was not a dict, defaulting to empty dict", logging.WARN)
        if "LOCAL_DIR" in all_elements.keys():
            self._local_dir = FileIndexingSchema._parseLocalDir(all_elements["LOCAL_DIR"])
        else:
            self._local_dir = Path("./data/")
            Logger.Log(f"{name} config does not have a 'LOCAL_DIR' element; defaulting to local_dir={self._local_dir}", logging.WARN)
        if "REMOTE_URL" in all_elements.keys():
            self._remote_url = FileIndexingSchema._parseRemoteURL(all_elements["REMOTE_URL"])
        else:
            self._remote_url = None
            Logger.Log(f"{name} config does not have a 'REMOTE_URL' element; defaulting to remote_url={self._remote_url}", logging.WARN)
        if "TEMPLATES_URL" in all_elements.keys():
            self._templates_url = FileIndexingSchema._parseTemplatesURL(all_elements["TEMPLATES_URL"])
        else:
            self._templates_url = "https://github.com/opengamedata/opengamedata-samples"
            Logger.Log(f"{name} config does not have a 'TEMPLATES_URL' element; defaulting to templates_url={self._templates_url}", logging.WARN)

        _used = {"LOCAL_DIR", "REMOTE_URL", "TEMPLATES_URL"}
        _leftovers = { key : val for key,val in all_elements.items() if key not in _used }
        super().__init__(name=name, other_elements=_leftovers)

    @property
    def LocalDirectory(self) -> Path:
        return self._local_dir

    @property
    def RemoteURL(self) -> Optional[str]:
        return self._remote_url

    @property
    def TemplatesURL(self) -> str:
        return self._templates_url

    @property
    def AsMarkdown(self) -> str:
        ret_val : str

        ret_val = f"{self.Name} : Local=_{self.LocalDirectory}_, Remote=_{self.RemoteURL}_"
        return ret_val

    @staticmethod
    def _parseLocalDir(dir) -> Path:
        ret_val : Path
        if isinstance(dir, Path):
            ret_val = dir
        elif isinstance(dir, str):
            ret_val = Path(dir)
        else:
            ret_val = Path(str(dir))
            Logger.Log(f"File Indexing local data directory was unexpected type {type(dir)}, defaulting to Path(str(dir))={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseRemoteURL(url) -> str:
        ret_val : str
        if isinstance(url, str):
            ret_val = url
        else:
            ret_val = str(url)
            Logger.Log(f"File indexing remote url was unexpected type {type(url)}, defaulting to str(url)={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseTemplatesURL(url) -> str:
        ret_val : str
        if isinstance(url, str):
            ret_val = url
        else:
            ret_val = str(url)
            Logger.Log(f"File indexing remote url was unexpected type {type(url)}, defaulting to str(url)={ret_val}.", logging.WARN)
        return ret_val



from typing import List, Optional, Dict, Any
from models.SequenceModel import SequenceModel
import datetime


## @class RecentPurchasesModel
# Returns a list of 8 lists, corresponding to the 8 items that can be purchased: (in this order) [homes, food,
# farms, fertilizer, livestock, skimmers, signs, roads].
# In each list is the number of seconds since that item was purchased, for each time it was purchased within the
# 10 most recent purchases.
# Rationale: This gives us options for displaying the data however we feel is most helpful. For example, we could
# just indicate the length of each list to indicate the number of purchases of each type within the last ten overall
# purchases, or we could incorporate the time stamps as well.
# @param levels: Levels applicable for model
class RecentPurchasesModel(SequenceModel):
    def __init__(self, levels: List[int] = []):
        '''
        @class RecentPurchasesModel
        Returns a list of 8 lists, corresponding to the 8 items that can be purchased: (in this order) [homes, food,
        farms, fertilizer, livestock, skimmers, signs, roads].
        In each list is the number of seconds since that item was purchased, for each time it was purchased within the
        10 most recent purchases.
        :param levels: Levels applicable for model
        '''
        super().__init__()

    def _eval(self, events: List[Dict[str, Any]], verbose: bool = False) -> List[List[int]]:
        assert events
        homes = []
        food = []
        farms = []
        fertilizer = []
        livestock = []
        skimmers = []
        signs = []
        roads = []
        count = 0
        now = events[-1]["client_time"]
        if type(now) is str:
            now = datetime.datetime.fromisoformat(now)
        for event in reversed(events):
            # gamestate
            if event["event_custom"] == 7 and event["event_data_complex"]["success"]:
                if event["event_data_complex"]["buy"] == 1:
                    event_time = event["client_time"]
                    if type(event_time) is str:
                        event_time = datetime.datetime.fromisoformat(event_time)
                    homes.append((now - event_time).total_seconds())
                elif event ["event_data_complex"]["buy"] == 2:
                    event_time = event["client_time"]
                    if type(event_time) is str:
                        event_time = datetime.datetime.fromisoformat(event_time)
                    food.append((now - event_time).total_seconds())
                elif event["event_data_complex"]["buy"] == 3:
                    event_time = event["client_time"]
                    if type(event_time) is str:
                        event_time = datetime.datetime.fromisoformat(event_time)
                    farms.append((now - event_time).total_seconds())
                elif event["event_data_complex"]["buy"] == 4:
                    event_time = event["client_time"]
                    if type(event_time) is str:
                        event_time = datetime.datetime.fromisoformat(event_time)
                    fertilizer.append((now - event_time).total_seconds())
                elif event["event_data_complex"]["buy"] == 5:
                    event_time = event["client_time"]
                    if type(event_time) is str:
                        event_time = datetime.datetime.fromisoformat(event_time)
                    livestock.append((now - event_time).total_seconds())
                elif event["event_data_complex"]["buy"] == 6:
                    event_time = event["client_time"]
                    if type(event_time) is str:
                        event_time = datetime.datetime.fromisoformat(event_time)
                    skimmers.append((now - event_time).total_seconds())
                elif event["event_data_complex"]["buy"] == 7:
                    event_time = event["client_time"]
                    if type(event_time) is str:
                        event_time = datetime.datetime.fromisoformat(event_time)
                    signs.append((now - event_time).total_seconds())
                elif event["event_data_complex"]["buy"] == 8:
                    event_time = event["client_time"]
                    if type(event_time) is str:
                        event_time = datetime.datetime.fromisoformat(event_time)
                    roads.append((now - event_time).total_seconds())
                # Check if ten purchases have been tallied. If so, break from the loop.
                count += 1
                if count >= 10:
                    break
        return [homes, food, farms, fertilizer, livestock, skimmers, signs, roads]



def __repr__(self):
        return f"RecentPurchasesModel(levels={self._levels}, input_type={self._input_type})"


## import standard libraries
import abc
import logging
from typing import Any, Dict, List, Optional
from ogd.core.generators.GeneratorLoader import GeneratorLoader
## import local files
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import Generator
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.models.enums.IterationMode import IterationMode

## @class Extractor
#  Abstract base class for game feature extractors.
#  Gives a few static functions to be used across all extractor classes,
#  and defines an interface that the SessionProcessor can use.
class GeneratorRegistry(abc.ABC):
    """Class for registering features to listen for events.

    :return: [description]
    :rtype: [type]
    """
    class Listener:
        def __init__(self, name:str, mode:IterationMode):
            self.name = name
            self.mode = mode
        
        def __str__(self) -> str:
            return f"{self.name} ({self.mode.name})"

        def __repr__(self) -> str:
            return str(self)

    # *** ABSTRACTS ***

    @abc.abstractmethod
    def _register(self, extractor:Generator, iter_mode:IterationMode):
        pass

    @abc.abstractmethod
    def _getGeneratorNames(self) -> List[str]:
        pass

    @abc.abstractmethod
    def _loadFromSchema(self, schema:GameSchema, loader:GeneratorLoader, overrides:Optional[List[str]]):
        pass

    @abc.abstractmethod
    def _updateFromEvent(self, event:Event) -> None:
        pass

    @abc.abstractmethod
    def _updateFromFeatureData(self, feature:FeatureData) -> None:
        pass

    # *** BUILT-INS & PROPERTIES ***

    # Base constructor for Registry.
    def __init__(self, mode:ExtractionMode):
        """Base constructor for Registry

        Just sets up mostly-empty dictionaries for use by the registry.
        _features is a list of feature orders, where each element is a map from feature names to actual Feature instances.
        _event_registry maps event names to Listener objects, which basically just say which feature(s) wants the given enent.
        _feature_registry maps feature names to Listener objects, which basically just say which 2nd-order feature(s) wants the given 1st-order feature.
        """
        self._event_registry : Dict[str,List[GeneratorRegistry.Listener]] = {"all_events":[]}
        self._mode           : ExtractionMode = mode

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    def Register(self, extractor:Generator, iter_mode:IterationMode):
        self._register(extractor=extractor, iter_mode=iter_mode)

    def GetGeneratorNames(self) -> List[str]:
        """Function to generate a list names of all enabled features, given a GameSchema
        This is different from the FeatureNames property of GameSchema,
        which ignores the 'enabled' attribute and does not expand per-count features
        (e.g. this function would include 'lvl0_someFeat', 'lvl1_someFeat', 'lvl2_someFeat', etc.
        while FeatureNames only would include 'someFeat').

        :param schema: The schema from which feature names should be generated.
        :type schema: GameSchema
        :return: A list of feature names.
        :rtype: List[str]
        """
        # TODO : Add error handling and/or timing and/or profiling
        return self._getGeneratorNames()

    def LoadFromSchema(self, schema:GameSchema, loader:GeneratorLoader, overrides:Optional[List[str]]):
        # TODO : Add error handling and/or timing and/or profiling
        self._loadFromSchema(schema=schema, loader=loader, overrides=overrides)

    def UpdateFromEvent(self, event:Event) -> None:
        """Perform extraction of features from a row.

        :param event: [description]
        :type event: Event
        :param table_schema: A data structure containing information on how the db
                             table assiciated with this game is structured.
        :type table_schema: TableSchema
        """
        # TODO : Add error handling and/or timing and/or profiling
        self._updateFromEvent(event=event)

    def UpdateFromFeatureData(self, feature:FeatureData) -> None:
        """Perform extraction of features from a row.

        :param event: [description]
        :type event: Event
        :param table_schema: A data structure containing information on how the db
                             table assiciated with this game is structured.
        :type table_schema: TableSchema
        """
        if isinstance(feature, FeatureData):
            self._updateFromFeatureData(feature=feature)
        else:
            Logger.Log(f"Got an invalid feature {feature} of type {type(feature)} for a registry in {self._mode.name} mode")

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***

    # def _format(obj):
    #     if obj == None:
    #         return ""
    #     elif type(obj) is timedelta:
    #         total_secs = obj.total_seconds()
    #         # h = total_secs // 3600
    #         # m = (total_secs % 3600) // 60
    #         # s = (total_secs % 3600) % 60 // 1  # just for reference
    #         # return f"{h:02.0f}:{m:02.0f}:{s:02.3f}"
    #         return str(total_secs)
    #     if obj is None:
    #         return ''
    #     else:
    #         return str(obj)



"""# import libraries
from datetime import timedelta
from typing import Any, Dict, List, Optional
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class AverageActiveTime(Feature):
    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)
        self.active_time_per_session: Dict[str, timedelta] = {} 
        self.session_count: Dict[str, int] = {}  

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["session_start", "pause_game", "unpause_game"]

    @classmethod
    def _featureFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["ActiveTime", "NumberOfSessionsPerPlayer"]  

    def _updateFromEvent(self, event: Event) -> None:
        if event.EventName == "session_start":
            player_id = event.PlayerID
            self.session_count[player_id] = self.session_count.get(player_id, 0) + 1
        elif event.EventName == "pause_game":
            player_id = event.PlayerID
            self.active_time_per_session[player_id] = self.active_time_per_session.get(player_id, timedelta())
            self.active_time_per_session[player_id] -= event.Timestamp
        elif event.EventName == "unpause_game":
            player_id = event.PlayerID
            self.active_time_per_session[player_id] = self.active_time_per_session.get(player_id, timedelta())
            self.active_time_per_session[player_id] += event.Timestamp

    def _updateFromFeatureData(self, feature: FeatureData):
        if feature.Name == "ActiveTime":
            self.active_time_per_session = feature.Value 
        elif feature.Name == "NumberOfSessionsPerPlayer":
            self.session_count = feature.Value 

    def _getFeatureValues(self) -> List[Any]:
        average_active_time = {}
        for player_id, active_time in self.active_time_per_session.items():
            session_count = self.session_count.get(player_id, 0)
            if session_count != 0:
                average_active_time[player_id] = active_time / session_count
        return [average_active_time]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"
"""

from typing import Any, Dict, List, Optional
from datetime import timedelta
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class AverageActiveTime(Feature):
    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)
        self.active_time_per_session: Dict[str, timedelta] = {}
        self.session_start_time: Dict[str, Optional[Event]] = {}

    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["session_start", "pause_game", "unpause_game"]

    @classmethod
    def _featureFilter(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event: Event) -> None:
        session_id = event.SessionID
        if event.EventName == "session_start":
            self.session_start_time[session_id] = event
        elif event.EventName == "pause_game":
            start_time = self.session_start_time.get(session_id)
            if start_time:
                active_time = event.Timestamp - start_time.Timestamp
                self.active_time_per_session[session_id] = self.active_time_per_session.get(session_id, timedelta()) + active_time
        elif event.EventName == "unpause_game":
            self.session_start_time[session_id] = event

    def _updateFromFeatureData(self, feature: FeatureData):
        pass

    def _getFeatureValues(self) -> List[Any]:
        total_active_time = sum(self.active_time_per_session.values(), timedelta())
        session_count = len(self.active_time_per_session)
        if session_count == 0:
            return [timedelta()]  # Return zero if no sessions
        average_active_time = total_active_time / session_count
        return [average_active_time]

    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"



# import standard libraries
import logging
import traceback
from typing import List, Dict, Type, Optional, Set
# import local files
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.registries.ExtractorRegistry import ExtractorRegistry
from ogd.core.processors.ExtractorProcessor import ExtractorProcessor
from ogd.core.processors.SessionProcessor import SessionProcessor
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExportMode import ExportMode
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.utils.Logger import Logger
from ogd.core.utils.utils import ExportRow

## @class PlayerProcessor
#  Class to extract and manage features for a processed csv file.
class PlayerProcessor(ExtractorProcessor):

    # *** BUILT-INS & PROPERTIES ***

    ## Constructor for the PlayerProcessor class.
    def __init__(self, LoaderClass: Type[GeneratorLoader], game_schema: GameSchema, player_id:str,
                 feature_overrides:Optional[List[str]]=None):
        """Constructor for the PlayerProcessor class.
           Simply stores some data for use later, including the type of extractor to use.

        :param LoaderClass: The type of data extractor to use for input data.
                            This should correspond to whatever game_id is in the TableSchema.
        :type LoaderClass: Type[GeneratorLoader]
        :param game_schema: A dictionary that defines how the game data itself is structured.
        :type game_schema: GameSchema
        :param player_id: _description_
        :type player_id: str
        :param feature_overrides: _description_, defaults to None
        :type feature_overrides: Union[List[str],None], optional
        :param player_file: _description_, defaults to None
        :type player_file: Optional[IO[str]], optional
        """
        Logger.Log(f"Setting up PlayerProcessor for {player_id}...", logging.DEBUG, depth=2)
        self._player_id : str      = player_id
        self._sessions  : Set[str] = set()
        super().__init__(LoaderClass=LoaderClass, game_schema=game_schema, feature_overrides=feature_overrides)
        ## Define instance vars
        Logger.Log(f"Done", logging.DEBUG, depth=2)

    def __str__(self):
        return f"PlayerProcessor({self._player_id})"

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @property
    def _mode(self) -> ExtractionMode:
        return ExtractionMode.PLAYER

    @property
    def _playerID(self) -> str:
        return self._player_id

    @property
    def _sessionID(self) -> str:
        return "player"

    def _getGeneratorNames(self) -> List[str]:
        if isinstance(self._registry, ExtractorRegistry):
            return ["PlayerID", "SessionCount"] + self._registry.GetGeneratorNames()
        else:
            raise TypeError()

    ## Function to handle processing of a single row of data.
    def _processEvent(self, event:Event):
        """Function to handle processing of a single row of data.
        Basically just responsible for ensuring an extractor for the session corresponding
        to the row already exists, then delegating the processing to that extractor.
        :param event: An object with the data for the event to be processed.
        :type event: Event
        """
        self._sessions.add(event.SessionID)
        self._registry.UpdateFromEvent(event=event)

    def _getLines(self) -> List[ExportRow]:
        ret_val : ExportRow
        # if as_str:
        #     ret_val = [self._player_id, len(self._sessions)] + self._registry.GetFeatureStringValues()
        # else:
        ret_val = [self._player_id, len(self._sessions)] + self._registry.GetFeatureValues()
        return [ret_val]

    def _getFeatureData(self, order:int) -> List[FeatureData]:
        return self._registry.GetFeatureData(order=order, player_id=self._player_id)

    ##  Function to empty the list of lines stored by the PlayerProcessor.
    def _clearLines(self) -> None:
        """Function to empty the list of lines stored by the PlayerProcessor.
        This is helpful if we're processing a lot of data and want to avoid eating too much memory.
        """
        Logger.Log(f"Clearing features from PlayerProcessor for {self._player_id}.", logging.DEBUG, depth=2)
        self._registry = ExtractorRegistry(mode=self._mode)

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



# import libraries
from typing import Any, Dict, List, Optional
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class CountyUnlockCount(Feature):
    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)
        self.county_unlocks: Dict[str, int] = {}

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["county_unlocked"]

    @classmethod
    def _featureFilter(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event: Event) -> None:
        county_name = event.EventData.get("county_name", None)
        if county_name:
            if county_name not in self.county_unlocks:
                self.county_unlocks[county_name] = 1
            else:
                self.county_unlocks[county_name] += 1

    def _updateFromFeatureData(self, feature: FeatureData):
        pass

    def _getFeatureValues(self) -> List[Any]:
        total_county_unlocks = sum(self.county_unlocks.values())
        return [total_county_unlocks, self.county_unlocks]

    def Subfeatures(self) -> List[str]:
        return ["Breakdown"]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"



# import standard libraries
import logging
import math
import re
from typing import Any, Dict, Optional, List

# import 3rd-party libraries

# import OGD libraries
from ogd.core.utils.Logger import Logger

# import locals

class SemanticVersion:

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, major:Optional[int], minor:Optional[int]=None, patch:Optional[int]=None, suffix:Optional[str]=None, suffix_ver:Optional[int]=None, fallback:Optional[str]=None):
        """
        Constructor for a SemanticVersion.

        Any parameter may be set to None; if nothing at all is given, the version will be treated as 0.0.0.

        :param major: The major version number
        :type major: Optional[int]
        :param minor: The minor version number, defaults to None
        :type minor: Optional[int], optional
        :param patch: The patch version number, defaults to None
        :type patch: Optional[int], optional
        :param suffix: An optional suffix for the version, such as 'alpha' or 'beta', defaults to None
        :type suffix: Optional[str], optional
        :param suffix_ver: An optional suffix version, allowing for overall suffixes such as 'alpha1' or 'beta2', defaults to None
        :type suffix_ver: Optional[int], optional
        :param fallback: _description_, defaults to None
        :type fallback: Optional[str], optional
        """
        self._major      : Optional[int] = max(0, major) if major is not None else None
        self._minor      : Optional[int] = max(0, minor) if minor is not None else None
        self._patch      : Optional[int] = max(0, patch) if patch is not None else None
        self._suffix     : Optional[str] = suffix
        self._suffix_ver : Optional[int] = max(0, suffix_ver) if suffix_ver is not None else None
        self._fallback   : Optional[str] = fallback

    def __repr__(self) -> str:
        return f"SemanticVersion : {self}"

    def __str__(self) -> str:
        _major      = str(self._major)      if self._major      is not None else None
        _minor      = f".{self._minor}"     if self._minor      is not None else None
        _patch      = f".{self._patch}"     if self._patch      is not None else None
        _suffix     = f"-{self._suffix}"    if self._suffix     is not None else None
        _suffix_ver = f"{self._suffix_ver}" if self._suffix_ver is not None else None

        return f"{_major}{_minor or ''}{_patch or ''}{_suffix or ''}{_suffix_ver or ''}" if self.IsValid else str(self._fallback)

    def __eq__(self, RHS:Any) -> bool:
        """
        _summary_

        :param RHS: _description_
        :type RHS: Any
        :raises TypeError: _description_
        :raises TypeError: _description_
        :raises TypeError: _description_
        :return: _description_
        :rtype: bool
        """
    # 1. Handle cases where RHS is not a SemanticVersion.
        if not isinstance(RHS, SemanticVersion):
            if isinstance(RHS, str):
                RHS = SemanticVersion.FromString(RHS, verbose=False)
            elif isinstance(RHS, int):
                RHS = SemanticVersion(major=RHS)
            else:
                return False
        if self.IsValid and RHS.IsValid:
    # 2. Handle general case, where we're just comparing each piece
            return  self._CompMajor  == RHS._CompMajor  \
                and self._CompMinor  == RHS._CompMinor  \
                and self._CompPatch  == RHS._CompPatch  \
                and self._suffix     == RHS._suffix     \
                and self._suffix_ver == RHS._suffix_ver
        else:
            return str(self) == str(RHS)

    def __gt__(self, RHS:Any) -> bool:
    # 1. Handle cases where RHS is not a SemanticVersion, or self is not a valid SemVer string.
        # a. If we're not valid, we're not bigger
        if not self.IsValid:
            return False
        # b. If they're not SemVer, convert
        if not isinstance(RHS, SemanticVersion):
            if isinstance(RHS, str):
                RHS = SemanticVersion.FromString(RHS, verbose=False)
            elif isinstance(RHS, int):
                RHS = SemanticVersion(major=RHS)
            else:
                raise TypeError(f"'>' not supported between instances '{type(self)}' and '{type(RHS)}'")
        # c. If they're not valid, but we are, we're bigger
        if not RHS.IsValid:
            return True
    # 2. Handle general case, where we're just comparing each piece
        # a. If our major is bigger, we're bigger
        if self._CompMajor > RHS._CompMajor:
            return True
        # b. If majors are equal, compare minors
        elif self._CompMajor == RHS._CompMajor:
        # c. If our minor is bigger, we're bigger
            if self._CompMinor > RHS._CompMinor:
                return True
        # d. If minors are equal, compare patches
            elif self._CompMinor == RHS._CompMinor:
        # e. If our patch is bigger, we're bigger
                if self._CompPatch > RHS._CompPatch:
                    return True
        # f. Otherwise, we're not bigger.
        return False

    def __ge__(self, RHS:Any):
    # 1. Handle cases where RHS is not a SemanticVersion, or self is not a valid SemVer string.
        # a. If we're not valid, we're not bigger, so return whether we're equal
        if not self.IsValid:
            return self == RHS
        # b. If they're not SemVer, convert
        if not isinstance(RHS, SemanticVersion):
            if isinstance(RHS, str):
                RHS = SemanticVersion.FromString(RHS, verbose=False)
            elif isinstance(RHS, int):
                RHS = SemanticVersion(major=RHS)
            else:
                raise TypeError(f"'>=' not supported between instances '{type(self)}' and '{type(RHS)}'")
        # c. If they're not valid, but we are, we're bigger
        if not RHS.IsValid:
            return True
    # 2. Handle general case, where we're just comparing each piece
        # a. If our major is bigger, we're bigger
        if self._CompMajor > RHS._CompMajor:
            return True
        # b. If majors are equal, compare minors
        elif self._CompMajor == RHS._CompMajor:
        # c. If our minor is bigger, we're bigger
            if self._CompMinor > RHS._CompMinor:
                return True
        # d. If minors are equal, compare patches
            elif self._CompMinor == RHS._CompMinor:
        # e. If our patch is bigger, we're bigger
                if self._CompPatch > RHS._CompPatch:
                    return True
        # d. If patches are equal, check if we're totally equal.
                elif self._CompPatch == RHS._CompPatch:
                    return self._suffix == RHS._suffix and self._suffix_ver == RHS._suffix_ver
        # f. Otherwise, we're not bigger.
        return False

    def __lt__(self, RHS:Any) -> bool:
    # 1. Handle cases where RHS is not a SemanticVersion, or self is not a valid SemVer string.
        # a. If we're not valid, we're smaller
        if not self.IsValid:
            return True
        # b. If they're not SemVer, convert
        if not isinstance(RHS, SemanticVersion):
            if isinstance(RHS, str):
                RHS = SemanticVersion.FromString(RHS, verbose=False)
            elif isinstance(RHS, int):
                RHS = SemanticVersion(major=RHS)
            else:
                raise TypeError(f"'<' not supported between instances '{type(self)}' and '{type(RHS)}'")
        # c. If they're not valid, but we are, we're not smaller
        if not RHS.IsValid:
            return False
    # 2. Handle general case, where we're just comparing each piece
        # a. If our major is smaller, we're smaller
        if self._CompMajor < RHS._CompMajor:
            return True
        # b. If majors are equal, compare minors
        elif self._CompMajor == RHS._CompMajor:
        # c. If our minor is smaller, we're smaller
            if self._CompMinor < RHS._CompMinor:
                return True
        # d. If minors are equal, compare patches
            elif self._CompMinor == RHS._CompMinor:
        # e. If our patch is smaller, we're smaller
                if self._CompPatch < RHS._CompPatch:
                    return True
        # f. Otherwise, we're not smaller.
        return False

    def __le__(self, RHS:Any):
    # 1. Handle cases where RHS is not a SemanticVersion, or self is not a valid SemVer string.
        # a. If we're not valid, we're smaller (or equal)
        if not self.IsValid:
            return True
        # b. If they're not SemVer, convert
        if not isinstance(RHS, SemanticVersion):
            if isinstance(RHS, str):
                RHS = SemanticVersion.FromString(RHS, verbose=False)
            elif isinstance(RHS, int):
                RHS = SemanticVersion(major=RHS)
            else:
                raise TypeError(f"'<=' not supported between instances '{type(self)}' and '{type(RHS)}'")
        # c. If they're not valid, but we are, we're not smaller
        if not RHS.IsValid:
            return False
    # 2. Handle general case, where we're just comparing each piece
        # a. If our major is smaller, we're smaller
        if self._CompMajor < RHS._CompMajor:
            return True
        # b. If majors are equal, compare minors
        elif self._CompMajor == RHS._CompMajor:
        # c. If our minor is smaller, we're smaller.
            if self._CompMinor < RHS._CompMinor:
                return True
        # d. If minors are equal, compare patches
            elif self._CompMinor == RHS._CompMinor:
        # e. If our patch is smaller, we're smaller
                if self._CompPatch < RHS._CompPatch:
                    return True
        # d. If patches are equal, check if we're totally equal.
                elif self._CompPatch == RHS._CompPatch:
                    return self._suffix == RHS._suffix and self._suffix_ver == RHS._suffix_ver
        # f. Otherwise, we're not bigger.
        return False

    # *** PUBLIC STATICS ***

    @staticmethod
    def FromString(semver:str, verbose:bool=True) -> 'SemanticVersion':
        pieces = re.split('\.|-', semver)
        if verbose:
            Logger.Log(f"Pieces: {pieces}", logging.DEBUG)

        return SemanticVersion._parseMajor(semver=semver, pieces=pieces, verbose=verbose)

    # *** PUBLIC METHODS ***

    # *** PROPERTIES ***

    @property
    def IsValid(self) -> bool:
        return self._major is not None

    @property
    def _CompMajor(self) -> int:
        return self._major or 0

    @property
    def _CompMinor(self) -> int:
        return self._minor or 0

    @property
    def _CompPatch(self) -> int:
        return self._patch or 0

    @staticmethod
    def _parseMajor(semver:str, pieces:List[str], verbose:bool) -> 'SemanticVersion':
        ret_val : 'SemanticVersion'

        _major : int
        try:
            _major = max(0, int(pieces[0]))
        except ValueError as err:
            _default_ver = semver
            if verbose:
                Logger.Log(f"Could not parse semantic version from {semver}, {pieces[0]} was not an int! Defaulting to version={_default_ver}\nError Details: {err}", level=logging.WARN)
            ret_val = SemanticVersion(major=None, fallback=semver)
        else:
            if len(pieces) > 1:
                ret_val = SemanticVersion._parseMinor(semver=semver, pieces=pieces[1:], major=_major, verbose=verbose)
            else:
                ret_val = SemanticVersion(major=_major)
        finally:
            return ret_val
                

    @staticmethod
    def _parseMinor(semver:str, pieces:List[str], major:int, verbose:bool) -> 'SemanticVersion':
        ret_val : 'SemanticVersion'

        _minor : int
        try:
            _minor = max(0, int(pieces[0]))
        except ValueError as err:
            _default_ver = str(major)
            if verbose:
                Logger.Log(f"Could not parse semantic version from {semver}, {pieces[0]} was not an int! Defaulting to version={_default_ver}\nError Details: {err}", level=logging.WARN)
            ret_val = SemanticVersion(major=major)
        else:
            if len(pieces) > 1:
                ret_val = SemanticVersion._parseFix(semver=semver, pieces=pieces[1:], major=major, minor=_minor, verbose=verbose)
            else:
                ret_val = SemanticVersion(major=major, minor=_minor)
        finally:
            return ret_val

    @staticmethod
    def _parseFix(semver:str, pieces:List[str], major:int, minor:int, verbose:bool) -> 'SemanticVersion':
        ret_val : 'SemanticVersion'

        _fix : int
        try:
            _fix = max(0, int(pieces[0]))
        except ValueError as err:
            _default_ver = f"{major}.{minor}"
            if verbose:
                Logger.Log(f"Could not parse semantic version from {semver}, {pieces[0]} was not an int! Defaulting to version={_default_ver}\nError Details: {err}", level=logging.WARN)
            ret_val = SemanticVersion(major=major, minor=minor)
        else:
            if len(pieces) > 1:
                ret_val = SemanticVersion._parseSuffix(semver=semver, pieces=pieces[1:], major=major, minor=minor, fix=_fix, verbose=verbose)
            else:
                ret_val = SemanticVersion(major=major, minor=minor, patch=_fix)
        finally:
            return ret_val

    @staticmethod
    def _parseSuffix(semver:str, pieces:List[str], major:int, minor:int, fix:int, verbose:bool) -> 'SemanticVersion':
        ret_val : 'SemanticVersion'

        _suffix : str
        _suffix = pieces[0]
        if len(pieces) > 1:
            ret_val = SemanticVersion._parseSuffixVersion(semver=semver, pieces=pieces[1:], major=major, minor=minor, fix=fix, suffix=_suffix, verbose=verbose)
        else:
            ret_val = SemanticVersion(major=major, minor=minor, patch=fix, suffix=_suffix)
        return ret_val

    @staticmethod
    def _parseSuffixVersion(semver:str, pieces:List[str], major:int, minor:int, fix:int, suffix:str, verbose:bool) -> 'SemanticVersion':
        ret_val : 'SemanticVersion'

        _suffix_version : int
        try:
            _suffix_version = max(0, int(pieces[0]))
        except ValueError as err:
            _default_ver = f"{major}.{minor}.{fix}-{suffix}"
            if verbose:
                Logger.Log(f"Could not parse semantic version from {semver}, {pieces[0]} was not an int! Defaulting to version={_default_ver}\nError Details: {err}", level=logging.WARN)
            ret_val = SemanticVersion(major=major, minor=minor, patch=fix, suffix=suffix)
        else:
            ret_val = SemanticVersion(major=major, minor=minor, patch=fix, suffix=suffix, suffix_ver=_suffix_version)
            if len(pieces) > 1 and verbose:
                Logger.Log(f"Semantic version string {semver} had excess parts, reducing to version={ret_val}", level=logging.WARN)
        finally:
            return ret_val


# import standard libraries
import logging
from typing import Any, Dict, List, Union
# import local files
from ogd.core.schemas.tables.ColumnSchema import ColumnSchema
from ogd.core.utils.Logger import Logger

class ColumnMapSchema:
    def __init__(self, map:Dict[str, Any], column_names:List[str]):
        self._map                  : Dict[str, Union[int, List[int], Dict[str,int], None]] = {
            "session_id"           : None,
            "app_id"               : None,
            "timestamp"            : None,
            "event_name"           : None,
            "event_data"           : None,
            "event_source"         : None,
            "app_version"          : None,
            "app_branch"           : None,
            "log_version"          : None,
            "time_offset"          : None,
            "user_id"              : None,
            "user_data"            : None,
            "game_state"           : None,
            "event_sequence_index" : None
        }
        self._other_elements       : Dict[str, Any]
        self._column_names = column_names

        if isinstance(map, dict):
            for key in self._map.keys():
                if key in map:
                    element = ColumnMapSchema._parseElement(elem=map[key], name=key)
                    if isinstance(element, str):
                        self._map[key] = column_names.index(element)
                    elif isinstance(element, list):
                        self._map[key] = [column_names.index(listelem) for listelem in element]
                    elif isinstance(element, dict):
                        self._map[key] = {key : column_names.index(listelem) for key,listelem in element.items()}
                else:
                    Logger.Log(f"Column config does not have a '{key}' element, defaulting to {key} : None", logging.WARN)
            self._other_elements = { key : val for key,val in map.items() if key not in self._map.keys() }
        else:
            self._other_elements = {}
            Logger.Log(f"For TableSchema config, map was not a dict, defaulting all items in map to None", logging.WARN)

    @property
    def Map(self) -> Dict[str, Union[int, List[int], Dict[str, int], None]]:
        """Mapping from Event element names to the indices of the database columns mapped to them.
        There may be a single index, indicating a 1-to-1 mapping of a database column to the element;
        There may be a list of indices, indicating multiple columns will be concatenated to form the element value;
        There may be a further mapping of keys to indicies, indicating multiple columns will be joined into a JSON object, with keys mapped to values found at the columns with given indices.

        :return: The dictionary mapping of element names to indices.
        :rtype: Dict[str, Union[int, List[int], Dict[str, int], None]]
        """
        return self._map

    @property
    def SessionID(self) -> Union[int, List[int], Dict[str, int], None]:
        return self._map['session_id']

    @property
    def AppID(self) -> Union[int, List[int], Dict[str, int], None]:
        return self._map['app_id']

    @property
    def Timestamp(self) -> Union[int, List[int], Dict[str, int], None]:
        return self._map['timestamp']

    @property
    def EventName(self) -> Union[int, List[int], Dict[str, int], None]:
        return self._map['event_name']

    @property
    def EventData(self) -> Union[int, List[int], Dict[str, int], None]:
        return self._map['event_data']

    @property
    def EventSource(self) -> Union[int, List[int], Dict[str, int], None]:
        return self._map['event_source']

    @property
    def AppVersion(self) -> Union[int, List[int], Dict[str, int], None]:
        return self._map['app_version']

    @property
    def AppBranch(self) -> Union[int, List[int], Dict[str, int], None]:
        return self._map['app_branch']

    @property
    def LogVersion(self) -> Union[int, List[int], Dict[str, int], None]:
        return self._map['log_version']

    @property
    def TimeOffset(self) -> Union[int, List[int], Dict[str, int], None]:
        return self._map['time_offset']

    @property
    def UserID(self) -> Union[int, List[int], Dict[str, int], None]:
        return self._map['user_id']

    @property
    def UserData(self) -> Union[int, List[int], Dict[str, int], None]:
        return self._map['user_data']

    @property
    def GameState(self) -> Union[int, List[int], Dict[str, int], None]:
        return self._map['game_state']

    @property
    def EventSequenceIndex(self) -> Union[int, List[int], Dict[str, int], None]:
        return self._map['event_sequence_index']

    @property
    def Elements(self) -> Dict[str, str]:
        return self._other_elements

    @property
    def ElementNames(self) -> List[str]:
        return list(self._other_elements.keys())

    @property
    def AsMarkdown(self) -> str:
        ret_val : str

        event_column_list = []
        for evt_col,row_col in self._map.items():
            if row_col is not None:
                if isinstance(row_col, str):
                    event_column_list.append(f"**{evt_col}** = Column '*{row_col}*'  ")
                elif isinstance(row_col, list):
                    mapped_list = ", ".join([f"'*{item}*'" for item in row_col])
                    event_column_list.append(f"**{evt_col}** = Columns {mapped_list}  ") # figure out how to do one string foreach item in list.
                elif isinstance(row_col, int):
                    event_column_list.append(f"**{evt_col}** = Column '*{self._column_names[row_col]}*' (index {row_col})  ")
                else:
                    event_column_list.append(f"**{evt_col}** = Column '*{row_col}*' (DEBUG: Type {type(row_col)})  ")
            else:
                event_column_list.append(f"**{evt_col}** = null  ")
        ret_val = "\n".join(event_column_list)
        return ret_val
    
    @staticmethod
    def _parseElement(elem:Any, name:str) -> Union[str, List[str], Dict[str, str], None]:
        ret_val : Union[str, List[str], Dict[str, str], None]
        if elem is not None:
            if isinstance(elem, str):
                ret_val = elem
            elif isinstance(elem, list):
                ret_val = elem
            elif isinstance(elem, dict):
                ret_val = elem
            else:
                ret_val = str(elem)
                Logger.Log(f"Column name(s) mapped to {name} was not a string or list, defaulting to str(name) == {ret_val} being mapped to {name}", logging.WARN)
        else:
            ret_val = None
            Logger.Log(f"Column name mapped to {name} was left null, nothing will be mapped to {name}", logging.WARN)
        return ret_val



# import libraries
from typing import Any, Dict, Final, List, Optional
from ogd.core.generators.Generator import GeneratorParameters
# import local files
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.models.Event import Event

class GameScript(SessionFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """

    TYPE : Final[Dict[int, str]] = {0: 'Dry', 1: 'No Humor', 2: 'No Snark', 3: 'Normal'}

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._name : Optional[str] = None
        self._version : Optional[int] = None

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["CUSTOM.1"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return [] 

    def _updateFromEvent(self, event:Event) -> None:
        if Event.CompareVersions(event.LogVersion, "7") >= 0:
            _type = event.EventData.get("script_type")
            self._version = event.EventData.get("script_version")
            if _type is not None:
                self._name = GameScript.TYPE.get(_type)
            else:
                raise ValueError(f"")

        return

    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._name, self._version]
    
    def Subfeatures(self) -> List[str]:
        return ["Version"]

    def BaseFeatureSuffix(self) -> str:
        return "Name"




# import libraries
import operator
import logging
from collections import Counter
from typing import Any, Final, List, Optional
# import local files
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.extractors.PerLevelFeature import PerLevelFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.utils.Logger import Logger




class QuitNode(SessionFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)

        self._quit_nodes : Counter = Counter()
        # >>> create/initialize any variables to track feature extractor state <<<
        #
        # e.g. To track whether extractor found a click event yet:
        # self._found_click : bool = False

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return [] # >>> fill in names of events this Feature should use for extraction. <<<

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["QuitLevel"] # >>> fill in names of first-order features this Feature should use for extraction. <<<

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        # >>> use the data in the Event object to update state variables as needed. <<<
        # Note that this function runs once on each Event whose name matches one of the strings returned by _eventFilter()
        #
        # e.g. check if the event name contains the substring "Click," and if so set self._found_click to True
        
        return

    def _updateFromFeatureData(self, feature: FeatureData):
        """_summary_

        :param feature: _description_
        :type feature: FeatureData
        """
        QNODE_INDEX : Final[int]    = 2 # The "Node" subfeature of the "QuitLevel" feature should be at index 2.
        quit_node   : Optional[str] = None
        if len(feature.FeatureValues) > QNODE_INDEX: # need to ensure there's at least 3 items, for index 2 to be valid...
            quit_node = str(feature.FeatureValues[2])
        else:
            Logger.Log("In QuitNode, no subfeature value found, defaulting to None!", logging.WARN)
        self._quit_nodes.update(quit_node)
        return

    def _getFeatureValues(self) -> List[Any]:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        ##return the top 5 highest values as a dictionary
        top_sorted : dict = { elem[0] : elem[1] for elem in self._quit_nodes.most_common(5) }

        return [top_sorted]


    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return [] # >>> fill in names of Subfeatures for which this Feature should extract values. <<<
    
    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.POPULATION] # >>> delete any modes you don't want run for your Feature. <<<
    
    # @staticmethod
    # def MinVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the minimum logging version for events to be processed by this Feature. <<<
    #     return super().MinVersion()

    # @staticmethod
    # def MaxVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the maximum logging version for events to be processed by this Feature. <<<
    #     return super().MaxVersion()



## @module
# Calculates number of checkpoint features in list achieved per 1000 seconds of gameplay time, given by time_feat.
# @param feature_list: list of features to use.
# @param levels: Levels to use this feature during
# @param time_feat: sessDuration" or "sess_time_active", depending if the model should be used for active time or
# overall time.


from typing import List, Optional
from models.FeatureModel import FeatureModel
from datetime import timedelta

## @class
# Calculates number of checkpoint features in list achieved per 1000 seconds of gameplay time, given by time_feat.
# @param feature_list: list of features to use.
# @param levels: Levels to use this feature during
# @param time_feat: sessDuration" or "sess_time_active", depending if the model should be used for active time or
# overall time.
class FeatVelocityModel(FeatureModel):
    def __init__(self, feature_list: List[str], levels: List[int] = [], time_feat='sess_time_active'):
        '''
        Calculates number of checkpoint features in list achieved per 1000 seconds of gameplay time, given by time_feat.
        :param feature_list: list of features to use.
        :param levels: Levels to use this feature during
        :param time_feat: sessDuration" or "sess_time_active", depending if the model should be used for active time or
        overall time.
        '''
        self._feat_list = feature_list
        self._time_feat = time_feat
        super().__init__()

    def _eval(self, sess: dict, verbose: bool = False) -> Optional[float]:
        if sess['continue'] or sess['debug']:
            return None
        time_to_vals = [sess[f] for f in self._feat_list]
        if not any(time_to_vals):
            return 0
        num_reached_feats = sum(bool(v) for v in time_to_vals)
        if all(time_to_vals):
            time = max(time_to_vals)
        else:
            time = sess[self._time_feat]
        if type(time) is timedelta:  # sessions features give float, but cgi might give timedelta
            time = time.total_seconds()

        return num_reached_feats / time * 1000


    def __repr__(self):
        return f"FeatVelocityModel(feature_list={self._feat_list}, time_feat='{self._time_feat}'" \
               f", levels={self._levels}, input_type={self._input_type})"



from typing import Any, Dict, List, Optional
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class BuildingUnlockCount(Feature):
    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)
        self.building_unlocks: Dict[str, int] = {}

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["unlock_building_type"]

    @classmethod
    def _featureFilter(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event: Event) -> None:
        building_type = event.EventData.get("building_type", None)
        if building_type:
            if building_type not in self.building_unlocks:
                self.building_unlocks[building_type] = 1
            else:
                self.building_unlocks[building_type] += 1

    def _updateFromFeatureData(self, feature: FeatureData):
        pass

    def _getFeatureValues(self) -> List[Any]:
        total_building_unlocks = sum(self.building_unlocks.values())
        return [total_building_unlocks, self.building_unlocks]

    def Subfeatures(self) -> List[str]:
        return ["Breakdown"]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"



"""Initializer for Aqualab detectors"""

__all__ = [
    "CollectFactNoJob",
    "DiveSiteNoEvidence",
    "EchoRoomChange",
    "Idle",
    "SceneChangeFrequently",
    "HintAndLeave",
    "TwoHints"
]

from . import CollectFactNoJob
from . import DiveSiteNoEvidence
from . import EchoRoomChange
from . import Idle
from . import HintAndLeave
from . import SceneChangeFrequently
from . import TwoHints



"""Initializer for Journalism features"""

__all__ = [
    "TextClickCount",
    "ChoiceClickCount",
    "SessionPlayTime",
    "SnippetReceivedCount",
    "MeanSnippetTime",
    "StoryCompleteTime",
    "SkillSequenceCount",
    "TotalLevelTime",
    "PlayerAttributes",
    "FinalAttributes",
    "StoryAlignment",
    "QuitLevel",
    "QuitType",
    "WorstAttribute",
    "WorstPlayerAttribute",
    "TopAttribute",
    "TopPlayerAttribute",
    "MaxedPlayerAttribute",
    "LevelCompleteCount",
    "LevelCompleted",
    "TotalFails",
    "FailureAttributes",
    "FailureCount",
    "ContinuesOnFail",
    "PlayTime",
    "UserPlayTime",
    "GameComplete",
    "QuitNode",
    "SnippetReplace",
    "StoryEditorTime",
    "AttributeView",
    "EditorNoteOpen",
    "TopPlayerQuitType",
    "StoryScore",
    "StoryAlignmentSequence", 
    "StoryScoreSequence",
    "LevelTime",
    "SnippetsCollected", 
    "SnippetsSubmitted"


]

from . import TextClickCount
from . import StoryEditorTime
from . import ChoiceClickCount
from . import SessionPlayTime
from . import SnippetReceivedCount
from . import MeanSnippetTime
from . import StoryCompleteTime
from . import SkillSequenceCount
from . import TotalLevelTime
from . import FinalAttributes
from . import PlayerAttributes
from . import StoryAlignment
from . import QuitLevel
from . import QuitType
from . import WorstAttribute
from . import WorstPlayerAttribute
from . import TopAttribute
from . import TopPlayerAttribute
from . import MaxedPlayerAttribute
from . import LevelCompleteCount
from . import LevelCompleted
from . import TotalFails
from . import FailureAttributes
from . import FailureCount
from . import ContinuesOnFail
from . import PlayTime
from . import UserPlayTime
from . import GameComplete
from . import QuitNode
from . import SnippetReplace
from . import AttributeView
from . import EditorNoteOpen
from . import TopPlayerQuitType
from . import StoryScore
from . import StoryAlignmentSequence
from . import StoryScoreSequence
from . import LevelTime
from . import SnippetsCollected
from . import SnippetsSubmitted



# import libraries
import logging
from typing import Optional, List
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.models.Event import Event

class PerCountyFeature(PerCountFeature):
    COUNTY_LIST = ["Hillside", "Forest", "Prairie", "Wetland", "Urban"]

    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    def _validateEventCountIndex(self, event: Event):
        ret_val: bool = False

        county_name = event.GameState.get('current_county', event.EventData.get('current_county', "COUNTY NAME NOT FOUND"))
        if county_name != "COUNTY NAME NOT FOUND":
            try:
                if self.COUNTY_LIST.index(county_name) == self.CountIndex:
                    ret_val = True
            except ValueError:
                self.WarningMessage(f"County name {county_name} not found in COUNTY_LIST in {type(self).__name__}, for event {event.EventName} with game state {event.GameState}")
        else:
            self.WarningMessage(f"Got invalid current_county name {county_name} for event {event.EventName} in {type(self).__name__}")

        return ret_val


# import standard libraries
import logging
from typing import Dict
# import local files
from ogd.core.schemas.games.DataElementSchema import DataElementSchema
from ogd.core.schemas.Schema import Schema
from ogd.core.utils.Logger import Logger

class GameStateSchema(Schema):
    """
    Dumb struct to contain a specification of a game's GameState in a GameSchema file.

    These essentially are just a set of elements in the GameState attribute of the game's Events.
    """
    def __init__(self, name:str, all_elements:Dict[str, Dict]):
        self._game_state  : Dict[str, DataElementSchema]

        if not isinstance(all_elements, dict):
            all_elements   = {}
            Logger.Log(f"For {name} Event config, all_elements was not a dict, defaulting to empty dict", logging.WARN)
        self._game_state = GameStateSchema._parseGameStateElements(event_data=all_elements)
        super().__init__(name=name, other_elements=None)

    @property
    def GameStateElements(self) -> Dict[str, DataElementSchema]:
        return self._game_state

    @property
    def AsMarkdown(self) -> str:
        return "\n\n".join([
            f"### **{self.Name}**",
            "\n".join(
                  [elem.AsMarkdown for elem in self.GameStateElements.values()]
                + (["- Other Elements:"] +
                   [f"  - **{elem_name}**: {elem_desc}" for elem_name,elem_desc in self.NonStandardElements]
                  ) if len(self.NonStandardElements) > 0 else []
            )
        ])

    @property
    def AsMarkdownTable(self) -> str:
        ret_val = [
            f"### **{self.Name}**",
            "#### Event Data",
            "\n".join(
                ["| **Name** | **Type** | **Description** | **Sub-Elements** |",
                 "| ---      | ---      | ---             | ---              |"]
              + [elem.AsMarkdownRow for elem in self.GameStateElements.values()]
            ),
        ]
        if len(self.NonStandardElements) > 0:
            ret_val.append("#### Other Elements")
            ret_val.append(
                "\n".join( [f"- **{elem_name}**: {elem_desc}  " for elem_name,elem_desc in self.NonStandardElements] )
            )
        return "\n\n".join(ret_val)

    @staticmethod
    def _parseGameStateElements(event_data):
        ret_val : Dict[str, DataElementSchema]
        if isinstance(event_data, dict):
            ret_val = {name:DataElementSchema(name=name, all_elements=elems) for name,elems in event_data.items()}
        else:
            ret_val = {}
            Logger.Log(f"event_data was unexpected type {type(event_data)}, defaulting to empty dict.", logging.WARN)
        return ret_val



import logging
import pandas as pd
from datetime import datetime
from pandas.io.parsers import TextFileReader
from pathlib import Path
from typing import Any, Dict, IO, List, Tuple, Optional
## import local files
from ogd.core.interfaces.EventInterface import EventInterface
from ogd.core.models.enums.IDMode import IDMode
from ogd.core.schemas.configs.GameSourceSchema import GameSourceSchema
from ogd.core.schemas.tables.TableSchema import TableSchema
from ogd.core.utils.Logger import Logger

class CSVInterface(EventInterface):

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, game_id:str, config:GameSourceSchema, fail_fast:bool, filepath:Path, delim:str = ','):
        # set up data from params
        self._filepath  : Path = filepath
        self._delimiter : str = delim
        super().__init__(game_id=game_id, config=config, fail_fast=fail_fast)
        # set up data from file
        self._data      : pd.DataFrame = pd.DataFrame()
        self.Open()

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    def _open(self) -> bool:
        try:
            # TODO should include option for access to the TableSchema in the interface, because obviously it should know what form the table takes.
            target_types = { 'session_id' : 'str' }
            _data = pd.read_csv(filepath_or_buffer=self._filepath, delimiter=self._delimiter, dtype=target_types, parse_dates=['timestamp'])
            # _data = pd.read_csv(filepath_or_buffer=self._filepath, delimiter=self._delimiter)
            Logger.Log(f"Loaded from CSV, columns are: {_data.dtypes}", logging.INFO)
            Logger.Log(f"First few rows are:\n{_data.head(n=3)}")
            self._data = _data.where(_data.notnull(), None)
            self._is_open = True
            return True
        except FileNotFoundError as err:
            Logger.Log(f"Could not find file {self._filepath}.", logging.ERROR)
            return False

    def _close(self) -> bool:
        self._is_open = False
        self._data = pd.DataFrame() # make new dataframe, let old data get garbage collected I assume.
        return True

    def _allIDs(self) -> List[str]:
        return [str(id) for id in self._data['session_id'].unique().tolist()]

    def _fullDateRange(self) -> Dict[str,datetime]:
        min_time = pd.to_datetime(self._data['timestamp'].min())
        max_time = pd.to_datetime(self._data['timestamp'].max())
        return {'min':min_time, 'max':max_time}

    def _rowsFromIDs(self, id_list: List[str], id_mode:IDMode=IDMode.SESSION, versions:Optional[List[int]]=None, exclude_rows:Optional[List[str]]=None) -> List[Tuple]:
        ret_val : List[Tuple] = []
        if self.IsOpen() and not self._data.empty:
            _data : pd.DataFrame
            if id_mode == IDMode.SESSION:
                _data = self._data.loc[self._data['session_id'].isin(id_list)]
            elif id_mode == IDMode.USER:
                _data = self._data.loc[self._data['user_id'].isin(id_list)]
            else:
                _data = self._data.loc[self._data['session_id'].isin(id_list)]
            if exclude_rows is not None:
                _mask = _data['event_name'].isin(exclude_rows)
                _data = _data[~_mask]
            ret_val = list(_data.itertuples(index=False, name=None))
        return ret_val

    def _IDsFromDates(self, min:datetime, max:datetime, versions:Optional[List[int]]=None) -> List[str]:
        if not self._data.empty:
            server_times = pd.to_datetime(self._data['server_time'])
            mask = (server_times >= pd.to_datetime(min)) & (server_times <= pd.to_datetime(max))
            if versions is not None and versions is not []:
                mask = mask & (self._data['app_version'].isin(versions))
            data_masked = self._data.loc[mask]
            return data_masked['session_id'].unique().tolist()
        else:
            return []

    def _datesFromIDs(self, id_list:List[str], id_mode:IDMode=IDMode.SESSION, versions:Optional[List[int]]=None) -> Dict[str, datetime]:
        if id_mode == IDMode.SESSION:
            min_date = self._data[self._data['session_id'].isin(id_list)]['timestamp'].min()
            max_date = self._data[self._data['session_id'].isin(id_list)]['timestamp'].max()
        elif id_mode == IDMode.USER:
            min_date = self._data[self._data['user_id'].isin(id_list)]['timestamp'].min()
            max_date = self._data[self._data['user_id'].isin(id_list)]['timestamp'].max()
        else:
            min_date = self._data[self._data['session_id'].isin(id_list)]['timestamp'].min()
            max_date = self._data[self._data['session_id'].isin(id_list)]['timestamp'].max()
        return {'min':pd.to_datetime(min_date), 'max':pd.to_datetime(max_date)}

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



# import libraries
from typing import Any, List
import json
# import locals
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData


orderMapping = {'1. One Box': 1, '2. Separated Boxes': 2, '3. Rotate a Pyramid': 3, '4. Match Silhouettes': 4, '5. Removing Objects': 5, '6. Stretch a Ramp': 6, '7. Max 2 Boxes': 7, '8. Combine 2 Ramps': 8, '9. Scaling Round Objects': 9,
                'Square Cross-Sections': 10, 'Bird Fez': 11, 'Pi Henge': 12, '45-Degree Rotations': 13,  'Pyramids are Strange': 14, 'Boxes Obscure Spheres': 15, 'Object Limits': 16, 'Warm Up': 17, 'Angled Silhouette': 18,
                'Sugar Cones': 19,'Stranger Shapes': 20, 'Tall and Small': 21, 'Ramp Up and Can It': 22, 'More Than Meets Your Eye': 23, 'Not Bird': 24, 'Unnecesary': 25, 'Zzz': 26, 'Bull Market': 27, 'Few Clues': 28, 'Orange Dance': 29, 'Bear Market': 30}
listPuzzles = list(orderMapping.keys())

class SequenceBetweenPuzzles(SessionFeature):
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._numPuzzles = 1
        self._currentPuzzle = {}
        self._activePuzzle = None
        self._userPuzzleDict = dict()

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["start_level", "puzzle_started", "create_shape", "check_solution", "puzzle_complete", "disconnect", "login_user", "exit_to_menu"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        ignoreEvent = False
        
        if event.EventName == "start_level":
            self._activePuzzle = event.EventData["task_id"]
            
            if self._activePuzzle == "Sandbox":
                ignoreEvent = True
            
            if ignoreEvent == False:
                if self._activePuzzle not in self._userPuzzleDict.keys():
                    self._userPuzzleDict[self._activePuzzle] = []
                self._currentPuzzle = {}
                self._currentPuzzle["sequence"] = self._numPuzzles
                self._currentPuzzle["funnel"] = "No data"

        if self._activePuzzle == "Sandbox":
            ignoreEvent = True
            
        if ignoreEvent == False:
        
            if event.EventName == "puzzle_started":
                self._currentPuzzle["funnel"] = "started"

            elif event.EventName == "create_shape":
                self._currentPuzzle["funnel"] = "shape_created"

            elif event.EventName == "check_solution":
                self._currentPuzzle["funnel"] = "submitted"

            elif event.EventName == "puzzle_complete":
                self._currentPuzzle["funnel"] = "completed"

            elif event.event_name in ["disconnect", "login_user", "exit_to_menu"] and self._activePuzzle != None:
                #Add current data
                self._userPuzzleDict[self._activePuzzle].append(json.dumps(self._currentPuzzle))
                self._currentPuzzle = {}
                self._numPuzzles += 1

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        listReturn = []
        listReturn.append(self._numPuzzles - 1)
        emptyList = []
        for puzzle in listPuzzles:
            if puzzle in self._userPuzzleDict.keys():
                listReturn.append(self._userPuzzleDict[puzzle])
            else:
                listReturn.append(emptyList)
        return listReturn
        
    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return listPuzzles



# import libraries
import logging
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.extractors.SessionFeature import SessionFeature

chime_dict = {'chime 1':0, 'chime 2':0, 'chime 3':0, 'chime 4':0, 'chime 5':0, 'chime 6':0, 'unknown chime':0}
class RingChimesCount(SessionFeature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._current_count : int = 0
        self._object_id = None
        self._chime_dict = chime_dict.copy()
        self._obj_list = list()

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["ring_chime"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        # self._current_count += 1
        self._object_id = event.event_data.get("note_played", "unknown chime")
        self._chime_dict[self._object_id] += 1
        
    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._chime_dict]

    